# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import pytest

import bugzilla

import tests
import tests.mockbackend


def test_api_component_edit():
    fakebz = tests.mockbackend.make_bz(
        component_create_args="data/mockargs/test_api_component_create1.txt",
        component_create_return={},
        component_update_args="data/mockargs/test_api_component_update1.txt",
        component_update_return={},
    )

    # addcomponent stub testing
    fakebz.addcomponent({
        "initialowner": "foo@example.com",
        "initialqacontact": "foo2@example.com",
        "initialcclist": "foo3@example.com",
        "product": "fooproduct",
        "is_active": 0,
    })

    # editcomponent stub testing
    fakebz.editcomponent({
        "initialowner": "foo@example.com",
        "blaharg": "blahval",
        "product": "fooproduct",
        "component": "foocomponent",
        "is_active": 0,
    })


def test_api_products():
    prod_list_return = {'ids': [1, 7]}
    prod_get_return = {'products': [
        {'id': 7, 'name': 'test-fake-product',
         'foo': {"bar": "baz"},
         'components': [
             {'default_assigned_to': 'Fake Guy',
              'name': 'client-interfaces'},
             {'default_assigned_to': 'ANother fake dude!',
              'name': 'configuration'},
         ]},
    ]}

    compnames = ["client-interfaces", "configuration"]
    fakebz = tests.mockbackend.make_bz(
        product_get_enterable_args=None,
        product_get_enterable_return=prod_list_return,
        product_get_selectable_args=None,
        product_get_selectable_return=prod_list_return,
        product_get_args="data/mockargs/test_api_products_get1.txt",
        product_get_return=prod_get_return,
    )

    # enterable products
    fakebz.product_get(ptype="enterable")
    fakebz.product_get(ptype="selectable")
    with pytest.raises(RuntimeError):
        fakebz.product_get(ptype="idontknow")

    # Double refresh things
    fakebz.getproducts(force_refresh=True, ptype="enterable")
    fakebz.getproducts(force_refresh=True, ptype="enterable")

    # getcomponents etc. testing
    fakebz = tests.mockbackend.make_bz(
        product_get_args="data/mockargs/test_api_products_get2.txt",
        product_get_return=prod_get_return,
    )

    # Lookup in product cache by name
    ret = fakebz.getcomponents("test-fake-product")
    assert ret == compnames
    # Lookup in product cache by id
    ret = fakebz.getcomponents(7)
    assert ret == compnames
    # force_refresh but its cool
    ret = fakebz.getcomponents("test-fake-product", force_refresh=True)
    assert ret == compnames

    # getcomponentsdetails usage
    fakebz = tests.mockbackend.make_bz(
        product_get_args="data/mockargs/test_api_products_get3.txt",
        product_get_return=prod_get_return,
    )
    fakebz.getcomponentdetails("test-fake-product", "configuration")

    # Some bit to test productget exclude_args
    fakebz = tests.mockbackend.make_bz(
        product_get_args="data/mockargs/test_api_products_get4.txt",
        product_get_return=prod_get_return)
    fakebz.product_get(ids=["7"], exclude_fields=["product.foo"])

    # Unknown product
    fakebz = tests.mockbackend.make_bz(
        product_get_args="data/mockargs/test_api_products_get5.txt",
        product_get_return=prod_get_return)
    with pytest.raises(bugzilla.BugzillaError):
        fakebz.getcomponents(0)



def test_bug_fields():
    fakebz = tests.mockbackend.make_bz(
        bug_fields_args="data/mockargs/test_bug_fields.txt",
        bug_fields_return="data/mockreturn/test_bug_fields.txt",
    )
    ret = fakebz.getbugfields(names=["bug_status"])
    assert ["bug_status"] == ret



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from logging import getLogger

import os
import sys
import urllib.parse

import requests


log = getLogger(__name__)


class _BugzillaSession(object):
    """
    Class to handle the backend agnostic 'requests' setup
    """
    def __init__(self, url, user_agent,
            sslverify, cert, tokencache, api_key,
            is_redhat_bugzilla,
            requests_session=None):
        self._url = url
        self._user_agent = user_agent
        self._scheme = urllib.parse.urlparse(url)[0]
        self._tokencache = tokencache
        self._api_key = api_key
        self._is_xmlrpc = False
        self._use_auth_bearer = False

        if self._scheme not in ["http", "https"]:
            raise Exception("Invalid URL scheme: %s (%s)" % (
                self._scheme, url))

        self._session = requests_session
        if not self._session:
            self._session = requests.Session()

        if cert:
            self._session.cert = cert
        if sslverify is False:
            self._session.verify = False
        self._session.headers["User-Agent"] = self._user_agent

        if is_redhat_bugzilla and self._api_key:
            self._use_auth_bearer = True
            self._session.headers["Authorization"] = (
                "Bearer %s" % self._api_key)

    def _get_timeout(self):
        # Default to 5 minutes. This is longer than bugzilla.redhat.com's
        # apparent 3 minute timeout so shouldn't affect legitimate usage,
        # but saves us from indefinite hangs
        DEFAULT_TIMEOUT = 300
        envtimeout = os.environ.get("PYTHONBUGZILLA_REQUESTS_TIMEOUT")
        return float(envtimeout or DEFAULT_TIMEOUT)

    def set_rest_defaults(self):
        self._session.headers["Content-Type"] = "application/json"
    def set_xmlrpc_defaults(self):
        self._is_xmlrpc = True
        self._session.headers["Content-Type"] = "text/xml"

    def get_user_agent(self):
        return self._user_agent
    def get_scheme(self):
        return self._scheme

    def get_auth_params(self):
        # bugzilla.redhat.com will error if there's auth bits in params
        # when Authorization header is used
        if self._use_auth_bearer:
            return {}

        # Don't add a token to the params list if an API key is set.
        # Keeping API key solo means bugzilla will definitely fail
        # if the key expires. Passing in a token could hide that
        # fact, which could make it confusing to pinpoint the issue.
        if self._api_key:
            # Bugzilla 5.0 only supports api_key as a query parameter.
            # Bugzilla 5.1+ takes it as a X-BUGZILLA-API-KEY header as well,
            # with query param taking preference.
            return {"Bugzilla_api_key": self._api_key}

        token = self._tokencache.get_value(self._url)
        if token:
            return {"Bugzilla_token": token}

        return {}

    def get_requests_session(self):
        return self._session

    def request(self, *args, **kwargs):
        timeout = self._get_timeout()
        if "timeout" not in kwargs:
            kwargs["timeout"] = timeout

        response = self._session.request(*args, **kwargs)

        if self._is_xmlrpc:
            # Yes this still appears to matter for properly decoding unicode
            # code points in bugzilla.redhat.com content
            response.encoding = "UTF-8"

        try:
            response.raise_for_status()
        except Exception as e:
            # Scrape the api key out of the returned exception string
            message = str(e).replace(self._api_key or "", "")
            raise type(e)(message).with_traceback(sys.exc_info()[2])

        return response



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import base64
import json
import logging
import os

from ._backendbase import _BackendBase
from .exceptions import BugzillaError
from ._util import listify


log = logging.getLogger(__name__)


def _update_key(indict, updict, key):
    if key not in indict:
        indict[key] = {}
    indict[key].update(updict.get(key, {}))


class _BackendREST(_BackendBase):
    """
    Internal interface for direct calls to bugzilla's REST API
    """
    def __init__(self, url, bugzillasession):
        _BackendBase.__init__(self, url, bugzillasession)
        self._bugzillasession.set_rest_defaults()


    #########################
    # Internal REST helpers #
    #########################

    def _handle_response(self, text):
        try:
            ret = dict(json.loads(text))
        except Exception:  # pragma: no cover
            log.debug("Failed to parse REST response. Output is:\n%s", text)
            raise

        if ret.get("error", False):
            raise BugzillaError(ret["message"], code=ret["code"])
        return ret

    def _op(self, method, apiurl, paramdict=None):
        fullurl = os.path.join(self._url, apiurl.lstrip("/"))
        log.debug("Bugzilla REST %s %s params=%s", method, fullurl, paramdict)

        data = None
        authparams = self._bugzillasession.get_auth_params()
        if method == "GET":
            authparams.update(paramdict or {})
        else:
            data = json.dumps(paramdict or {})

        response = self._bugzillasession.request(method, fullurl, data=data,
                params=authparams)
        return self._handle_response(response.text)

    def _get(self, *args, **kwargs):
        return self._op("GET", *args, **kwargs)
    def _put(self, *args, **kwargs):
        return self._op("PUT", *args, **kwargs)
    def _post(self, *args, **kwargs):
        return self._op("POST", *args, **kwargs)


    #######################
    # API implementations #
    #######################

    def get_xmlrpc_proxy(self):
        raise BugzillaError("You are using the bugzilla REST API, "
                "so raw XMLRPC access is not provided.")
    def is_rest(self):
        return True

    def bugzilla_version(self):
        return self._get("/version")

    def bug_create(self, paramdict):
        return self._post("/bug", paramdict)
    def bug_fields(self, paramdict):
        return self._get("/field/bug", paramdict)
    def bug_get(self, bug_ids, aliases, paramdict):
        data = paramdict.copy()
        data["id"] = listify(bug_ids)
        data["alias"] = listify(aliases)
        ret = self._get("/bug", data)
        return ret

    def bug_attachment_get(self, attachment_ids, paramdict):
        # XMLRPC supported mutiple fetch at once, but not REST
        ret = {}
        for attid in listify(attachment_ids):
            out = self._get("/bug/attachment/%s" % attid, paramdict)
            _update_key(ret, out, "attachments")
            _update_key(ret, out, "bugs")
        return ret

    def bug_attachment_get_all(self, bug_ids, paramdict):
        # XMLRPC supported mutiple fetch at once, but not REST
        ret = {}
        for bugid in listify(bug_ids):
            out = self._get("/bug/%s/attachment" % bugid, paramdict)
            _update_key(ret, out, "attachments")
            _update_key(ret, out, "bugs")
        return ret

    def bug_attachment_create(self, bug_ids, data, paramdict):
        if data is not None and "data" not in paramdict:
            paramdict["data"] = base64.b64encode(data).decode("utf-8")
        paramdict["ids"] = listify(bug_ids)
        return self._post("/bug/%s/attachment" % paramdict["ids"][0],
                paramdict)

    def bug_attachment_update(self, attachment_ids, paramdict):
        paramdict["ids"] = listify(attachment_ids)
        return self._put("/bug/attachment/%s" % paramdict["ids"][0], paramdict)

    def bug_comments(self, bug_ids, paramdict):
        # XMLRPC supported mutiple fetch at once, but not REST
        ret = {}
        for bugid in bug_ids:
            out = self._get("/bug/%s/comment" % bugid, paramdict)
            _update_key(ret, out, "bugs")
        return ret
    def bug_history(self, bug_ids, paramdict):
        # XMLRPC supported mutiple fetch at once, but not REST
        ret = {"bugs": []}
        for bugid in bug_ids:
            out = self._get("/bug/%s/history" % bugid, paramdict)
            ret["bugs"].extend(out.get("bugs", []))
        return ret

    def bug_search(self, paramdict):
        return self._get("/bug", paramdict)
    def bug_update(self, bug_ids, paramdict):
        data = paramdict.copy()
        data["ids"] = listify(bug_ids)
        return self._put("/bug/%s" % data["ids"][0], data)
    def bug_update_tags(self, bug_ids, paramdict):
        raise BugzillaError("No REST API available for bug_update_tags")

    def component_create(self, paramdict):
        return self._post("/component", paramdict)
    def component_update(self, paramdict):
        if "ids" in paramdict:
            apiurl = str(listify(paramdict["ids"])[0])  # pragma: no cover
        if "names" in paramdict:
            apiurl = ("%(product)s/%(component)s" %
                    listify(paramdict["names"])[0])
        return self._put("/component/%s" % apiurl, paramdict)

    def externalbugs_add(self, paramdict):  # pragma: no cover
        raise BugzillaError(
            "No REST API available yet for externalbugs_add")
    def externalbugs_remove(self, paramdict):  # pragma: no cover
        raise BugzillaError(
            "No REST API available yet for externalbugs_remove")
    def externalbugs_update(self, paramdict):  # pragma: no cover
        raise BugzillaError(
            "No REST API available yet for externalbugs_update")

    def group_get(self, paramdict):
        return self._get("/group", paramdict)

    def product_get(self, paramdict):
        return self._get("/product/get", paramdict)
    def product_get_accessible(self):
        return self._get("/product_accessible")
    def product_get_enterable(self):
        return self._get("/product_enterable")
    def product_get_selectable(self):
        return self._get("/product_selectable")

    def user_create(self, paramdict):
        return self._post("/user", paramdict)
    def user_get(self, paramdict):
        return self._get("/user", paramdict)
    def user_login(self, paramdict):
        return self._get("/login", paramdict)
    def user_logout(self):
        return self._get("/logout")
    def user_update(self, paramdict):
        urlid = None
        if "ids" in paramdict:
            urlid = listify(paramdict["ids"])[0]  # pragma: no cover
        if "names" in paramdict:
            urlid = listify(paramdict["names"])[0]
        return self._put("/user/%s" % urlid, paramdict)



#!/usr/bin/env python3

import glob
import os
import shutil
import subprocess
import sys

import setuptools


def get_version():
    f = open("bugzilla/apiversion.py")
    for line in f:
        if line.startswith('version = '):
            return eval(line.split('=')[-1])  # pylint: disable=eval-used


class PylintCommand(setuptools.Command):
    user_options = []

    def initialize_options(self):
        pass
    def finalize_options(self):
        pass

    def run(self):
        import pylint.lint
        import pycodestyle

        files = (["bugzilla-cli", "bugzilla", "setup.py"] +
            glob.glob("examples/*.py") +
            glob.glob("tests/*.py"))
        output_format = sys.stdout.isatty() and "colorized" or "text"

        print("running pycodestyle")
        style_guide = pycodestyle.StyleGuide(
            config_file='tox.ini',
            format="pylint",
            paths=files,
        )
        report = style_guide.check_files()
        if style_guide.options.count:
            sys.stderr.write(str(report.total_errors) + '\n')

        print("running pylint")
        pylint_opts = [
            "--rcfile", ".pylintrc",
            "--output-format=%s" % output_format,
        ]
        pylint.lint.Run(files + pylint_opts)


class RPMCommand(setuptools.Command):
    description = ("Build src and binary rpms and output them "
        "in the source directory")
    user_options = []

    def initialize_options(self):
        pass
    def finalize_options(self):
        pass

    def run(self):
        self.run_command('sdist')
        srcdir = os.path.dirname(__file__)
        cmd = [
            "rpmbuild", "-ta",
            "--define", "_rpmdir %s" % srcdir,
            "--define", "_srcrpmdir %s" % srcdir,
            "--define", "_specdir /tmp",
            "dist/python-bugzilla-%s.tar.gz" % get_version(),
        ]
        subprocess.check_call(cmd)


class ManCommand(setuptools.Command):
    description = ("Regenerate manpages from rst")
    user_options = []

    def initialize_options(self):
        pass
    def finalize_options(self):
        pass

    def _make_man_pages(self):
        rstbin = shutil.which("rst2man")
        if not rstbin:
            rstbin = shutil.which("rst2man.py")
        if not rstbin:
            raise RuntimeError("Didn't find rst2man or rst2man.py")

        for path in glob.glob("man/*.rst"):
            base = os.path.basename(path)
            appname = os.path.splitext(base)[0]
            newpath = os.path.join(os.path.dirname(path),
                                   appname + ".1")

            print("Generating %s" % newpath)
            out = subprocess.check_output([rstbin, path])
            open(newpath, "wb").write(out)

            self.distribution.data_files.append(
                ('share/man/man1', (newpath,)))

    def run(self):
        self._make_man_pages()


def _parse_requirements(fname):
    ret = []
    for line in open(fname).readlines():
        if not line or line.startswith("#"):
            continue
        ret.append(line)
    return ret


setuptools.setup(
    name='python-bugzilla',
    version=get_version(),
    description='Library and command line tool for interacting with Bugzilla',
    license="GPLv2",
    url='https://github.com/python-bugzilla/python-bugzilla',
    classifiers=[
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: '
        'GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    packages=['bugzilla'],
    data_files=[('share/man/man1', ['man/bugzilla.1'])],
    entry_points={'console_scripts': ['bugzilla = bugzilla._cli:cli']},

    install_requires=_parse_requirements("requirements.txt"),
    tests_require=_parse_requirements("test-requirements.txt"),

    cmdclass={
        "regenerate_manpages": ManCommand,
        "pylint": PylintCommand,
        "rpm": RPMCommand,
    },
)



# Copyright (C) 2007, 2008, 2009, 2010 Red Hat Inc.
# Author: Will Woods <wwoods@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import copy
from logging import getLogger


log = getLogger(__name__)


class Bug(object):
    """
    A container object for a bug report. Requires a Bugzilla instance -
    every Bug is on a Bugzilla, obviously.
    Optional keyword args:
        dict=DICT   - populate attributes with the result of a getBug() call
        bug_id=ID   - if dict does not contain bug_id, this is required before
                      you can read any attributes or make modifications to this
                      bug.
    """
    def __init__(self, bugzilla, bug_id=None, dict=None, autorefresh=False):
        # pylint: disable=redefined-builtin
        # API had pre-existing issue that we can't change ('dict' usage)

        self.bugzilla = bugzilla
        self._rawdata = {}
        self.autorefresh = autorefresh

        # pylint: disable=protected-access
        self._aliases = self.bugzilla._get_bug_aliases()
        # pylint: enable=protected-access

        if not dict:
            dict = {}
        if bug_id:
            dict["id"] = bug_id

        self._update_dict(dict)
        self.weburl = bugzilla.url.replace('xmlrpc.cgi',
                                           'show_bug.cgi?id=%i' % self.bug_id)

    def __str__(self):
        """
        Return a simple string representation of this bug
        """
        return self.__unicode__()

    def __unicode__(self):
        """
        Return a simple unicode string representation of this bug
        """
        return "#%-6s %-10s - %s - %s" % (self.bug_id, self.bug_status,
                                          self.assigned_to, self.summary)

    def __repr__(self):
        url = ""
        if self.bugzilla:
            url = self.bugzilla.url
        return '<Bug #%i on %s at %#x>' % (self.bug_id, url, id(self))

    def __getattr__(self, name):
        refreshed = False
        while True:
            if refreshed and name in self.__dict__:
                # If name was in __dict__ to begin with, __getattr__ would
                # have never been called.
                return self.__dict__[name]

            for newname, oldname in self._aliases:
                if name == oldname and newname in self.__dict__:
                    return self.__dict__[newname]

            # Doing dir(bugobj) does getattr __members__/__methods__,
            # don't refresh for those
            if name.startswith("__") and name.endswith("__"):
                break

            if refreshed or not self.autorefresh:
                break

            log.info("Bug %i missing attribute '%s' - doing implicit "
                "refresh(). This will be slow, if you want to avoid "
                "this, properly use query/getbug include_fields, and "
                "set bugzilla.bug_autorefresh = False to force failure.",
                self.bug_id, name)

            # We pass the attribute name to getbug, since for something like
            # 'attachments' which downloads lots of data we really want the
            # user to opt in.
            self.refresh(extra_fields=[name])
            refreshed = True

        msg = ("Bug object has no attribute '%s'." % name)
        if not self.autorefresh:
            msg += ("\nIf '%s' is a bugzilla attribute, it may not have "
                    "been cached when the bug was fetched. You may want "
                    "to adjust your include_fields for getbug/query." % name)
        raise AttributeError(msg)

    def get_raw_data(self):
        """
        Return the raw API dictionary data that has been used to
        populate this bug
        """
        return copy.deepcopy(self._rawdata)

    def refresh(self, include_fields=None, exclude_fields=None,
        extra_fields=None):
        """
        Refresh the bug with the latest data from bugzilla
        """
        # pylint: disable=protected-access
        extra_fields = list(self._rawdata.keys()) + (extra_fields or [])
        r = self.bugzilla._getbug(self.bug_id,
            include_fields=include_fields, exclude_fields=exclude_fields,
            extra_fields=extra_fields)
        # pylint: enable=protected-access
        self._update_dict(r)
    reload = refresh

    def _translate_dict(self, newdict):
        if self.bugzilla:
            self.bugzilla.post_translation({}, newdict)

        for newname, oldname in self._aliases:
            if oldname not in newdict:
                continue

            if newname not in newdict:
                newdict[newname] = newdict[oldname]
            elif newdict[newname] != newdict[oldname]:
                log.debug("Update dict contained differing alias values "
                          "d[%s]=%s and d[%s]=%s , dropping the value "
                          "d[%s]", newname, newdict[newname], oldname,
                        newdict[oldname], oldname)
            del(newdict[oldname])


    def _update_dict(self, newdict):
        """
        Update internal dictionary, in a way that ensures no duplicate
        entries are stored WRT field aliases
        """
        self._translate_dict(newdict)
        self._rawdata.update(newdict)
        self.__dict__.update(newdict)

        if 'id' not in self.__dict__ and 'bug_id' not in self.__dict__:
            raise TypeError("Bug object needs a bug_id")


    ##################
    # pickle helpers #
    ##################

    def __getstate__(self):
        ret = self._rawdata.copy()
        ret["_aliases"] = self._aliases
        return ret

    def __setstate__(self, vals):
        self._rawdata = {}
        self.bugzilla = None
        self._aliases = vals.get("_aliases", [])
        self.autorefresh = False
        self._update_dict(vals)


    #####################
    # Modify bug status #
    #####################

    def setstatus(self, status, comment=None, private=False):
        """
        Update the status for this bug report.
        Commonly-used values are ASSIGNED, MODIFIED, and NEEDINFO.

        To change bugs to CLOSED, use .close() instead.
        """
        # Note: fedora bodhi uses this function
        vals = self.bugzilla.build_update(status=status,
                                          comment=comment,
                                          comment_private=private)
        log.debug("setstatus: update=%s", vals)

        return self.bugzilla.update_bugs(self.bug_id, vals)

    def close(self, resolution, dupeid=None, fixedin=None,
              comment=None, isprivate=False):
        """
        Close this bug.
        Valid values for resolution are in bz.querydefaults['resolution_list']
        For bugzilla.redhat.com that's:
        ['NOTABUG', 'WONTFIX', 'DEFERRED', 'WORKSFORME', 'CURRENTRELEASE',
         'RAWHIDE', 'ERRATA', 'DUPLICATE', 'UPSTREAM', 'NEXTRELEASE',
         'CANTFIX', 'INSUFFICIENT_DATA']
        If using DUPLICATE, you need to set dupeid to the ID of the other bug.
        If using WORKSFORME/CURRENTRELEASE/RAWHIDE/ERRATA/UPSTREAM/NEXTRELEASE
          you can (and should) set 'new_fixed_in' to a string representing the
          version that fixes the bug.
        You can optionally add a comment while closing the bug. Set 'isprivate'
          to True if you want that comment to be private.
        """
        # Note: fedora bodhi uses this function
        vals = self.bugzilla.build_update(comment=comment,
                                          comment_private=isprivate,
                                          resolution=resolution,
                                          dupe_of=dupeid,
                                          fixed_in=fixedin,
                                          status=str("CLOSED"))
        log.debug("close: update=%s", vals)

        return self.bugzilla.update_bugs(self.bug_id, vals)


    #####################
    # Modify bug emails #
    #####################

    def setassignee(self, assigned_to=None,
                    qa_contact=None, comment=None):
        """
        Set any of the assigned_to or qa_contact fields to a new
        bugzilla account, with an optional comment, e.g.
        setassignee(assigned_to='wwoods@redhat.com')
        setassignee(qa_contact='wwoods@redhat.com', comment='wwoods QA ftw')

        You must set at least one of the two assignee fields, or this method
        will throw a ValueError.

        Returns [bug_id, mailresults].
        """
        if not (assigned_to or qa_contact):
            raise ValueError("You must set one of assigned_to "
                             " or qa_contact")

        vals = self.bugzilla.build_update(assigned_to=assigned_to,
                                          qa_contact=qa_contact,
                                          comment=comment)
        log.debug("setassignee: update=%s", vals)

        return self.bugzilla.update_bugs(self.bug_id, vals)

    def addcc(self, cclist, comment=None):
        """
        Adds the given email addresses to the CC list for this bug.
        cclist: list of email addresses (strings)
        comment: optional comment to add to the bug
        """
        vals = self.bugzilla.build_update(comment=comment,
                                          cc_add=cclist)
        log.debug("addcc: update=%s", vals)

        return self.bugzilla.update_bugs(self.bug_id, vals)

    def deletecc(self, cclist, comment=None):
        """
        Removes the given email addresses from the CC list for this bug.
        """
        vals = self.bugzilla.build_update(comment=comment,
                                          cc_remove=cclist)
        log.debug("deletecc: update=%s", vals)

        return self.bugzilla.update_bugs(self.bug_id, vals)


    ####################
    # comment handling #
    ####################

    def addcomment(self, comment, private=False):
        """
        Add the given comment to this bug. Set private to True to mark this
        comment as private.
        """
        # Note: fedora bodhi uses this function
        vals = self.bugzilla.build_update(comment=comment,
                                          comment_private=private)
        log.debug("addcomment: update=%s", vals)

        return self.bugzilla.update_bugs(self.bug_id, vals)

    def getcomments(self):
        """
        Returns an array of comment dictionaries for this bug
        """
        comment_list = self.bugzilla.get_comments([self.bug_id])
        return comment_list['bugs'][str(self.bug_id)]['comments']


    #####################
    # Get/Set bug flags #
    #####################

    def get_flag_type(self, name):
        """
        Return flag_type information for a specific flag

        Older RHBugzilla returned a lot more info here, but it was
        non-upstream and is now gone.
        """
        for t in self.flags:
            if t['name'] == name:
                return t
        return None

    def get_flags(self, name):
        """
        Return flag value information for a specific flag
        """
        ft = self.get_flag_type(name)
        if not ft:
            return None

        return [ft]

    def get_flag_status(self, name):
        """
        Return a flag 'status' field

        This method works only for simple flags that have only a 'status' field
        with no "requestee" info, and no multiple values. For more complex
        flags, use get_flags() to get extended flag value information.
        """
        f = self.get_flags(name)
        if not f:
            return None

        # This method works only for simple flags that have only one
        # value set.
        assert len(f) <= 1

        return f[0]['status']

    def updateflags(self, flags):
        """
        Thin wrapper around build_update(flags=X). This only handles simple
        status changes, anything like needinfo requestee needs to call
        build_update + update_bugs directly

        :param flags: Dictionary of the form {"flagname": "status"}, example
            {"needinfo": "?", "devel_ack": "+"}
        """
        flaglist = []
        for key, value in flags.items():
            flaglist.append({"name": key, "status": value})
        return self.bugzilla.update_bugs([self.bug_id],
            self.bugzilla.build_update(flags=flaglist))


    ########################
    # Experimental methods #
    ########################

    def get_attachments(self, include_fields=None, exclude_fields=None):
        """
        Helper call to Bugzilla.get_attachments. If you want to fetch
        specific attachment IDs, use that function instead
        """
        if "attachments" in self.__dict__:
            return self.attachments

        data = self.bugzilla.get_attachments([self.bug_id], None,
                include_fields, exclude_fields)
        return data["bugs"][str(self.bug_id)]

    def get_attachment_ids(self):
        """
        Helper function to return only the attachment IDs for this bug
        """
        return [a["id"] for a in self.get_attachments(exclude_fields=["data"])]

    def get_history_raw(self):
        """
        Experimental. Get the history of changes for this bug.
        """
        return self.bugzilla.bugs_history_raw([self.bug_id])


class User(object):
    """
    Container object for a bugzilla User.

    :arg bugzilla: Bugzilla instance that this User belongs to.
    Rest of the params come straight from User.get()
    """
    def __init__(self, bugzilla, **kwargs):
        self.bugzilla = bugzilla
        self.__userid = kwargs.get('id')
        self.__name = kwargs.get('name')

        self.__email = kwargs.get('email', self.__name)
        self.__can_login = kwargs.get('can_login', False)

        self.real_name = kwargs.get('real_name', None)
        self.password = None

        self.groups = kwargs.get('groups', {})
        self.groupnames = []
        for g in self.groups:
            if "name" in g:
                self.groupnames.append(g["name"])
        self.groupnames.sort()


    ########################
    # Read-only attributes #
    ########################

    # We make these properties so that the user cannot set them.  They are
    # unaffected by the update() method so it would be misleading to let them
    # be changed.
    @property
    def userid(self):
        return self.__userid

    @property
    def email(self):
        return self.__email

    @property
    def can_login(self):
        return self.__can_login

    # name is a key in some methods.  Mark it dirty when we change it #
    @property
    def name(self):
        return self.__name

    def refresh(self):
        """
        Update User object with latest info from bugzilla
        """
        newuser = self.bugzilla.getuser(self.email)
        self.__dict__.update(newuser.__dict__)

    def updateperms(self, action, groups):
        """
        A method to update the permissions (group membership) of a bugzilla
        user.

        :arg action: add, remove, or set
        :arg groups: list of groups to be added to (i.e. ['fedora_contrib'])
        """
        self.bugzilla.updateperms(self.name, action, groups)


class Group(object):
    """
    Container object for a bugzilla Group.

    :arg bugzilla: Bugzilla instance that this Group belongs to.
    Rest of the params come straight from Group.get()
    """
    def __init__(self, bugzilla, **kwargs):
        self.bugzilla = bugzilla
        self.__groupid = kwargs.get('id')

        self.name = kwargs.get('name')
        self.description = kwargs.get('description', self.name)
        self.is_active = kwargs.get('is_active', False)
        self.icon_url = kwargs.get('icon_url', None)
        self.is_active_bug_group = kwargs.get('is_active_bug_group', None)

        self.membership = kwargs.get('membership', [])
        self.__member_emails = set()
        self._refresh_member_emails_list()

    ########################
    # Read-only attributes #
    ########################

    # We make these properties so that the user cannot set them.  They are
    # unaffected by the update() method so it would be misleading to let them
    # be changed.
    @property
    def groupid(self):
        return self.__groupid

    @property
    def member_emails(self):
        return sorted(self.__member_emails)

    def _refresh_member_emails_list(self):
        """
        Refresh the list of emails of the members of the group.
        """
        if self.membership:
            for m in self.membership:
                if "email" in m:
                    self.__member_emails.add(m["email"])

    def refresh(self, membership=False):
        """
        Update Group object with latest info from bugzilla
        """
        newgroup = self.bugzilla.getgroup(
            self.name, membership=membership)
        self.__dict__.update(newgroup.__dict__)
        self._refresh_member_emails_list()

    def members(self):
        """
        Retrieve the members of this Group from bugzilla
        """
        if not self.membership:
            self.refresh(membership=True)
        return self.membership



# rhbugzilla.py - a Python interface to Red Hat Bugzilla using xmlrpclib.
#
# Copyright (C) 2008-2012 Red Hat Inc.
# Author: Will Woods <wwoods@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from logging import getLogger

from ._util import listify

log = getLogger(__name__)


class _RHBugzillaConverters(object):
    """
    Static class that holds functional Red Hat back compat converters.
    Called inline in Bugzilla
    """
    @staticmethod
    def convert_build_update(
            component=None,
            fixed_in=None,
            qa_whiteboard=None,
            devel_whiteboard=None,
            internal_whiteboard=None,
            sub_component=None):
        adddict = {}

        def get_alias():
            # RHBZ has a custom extension to allow a bug to have multiple
            # aliases, so the format of aliases is
            #    {"add": [...], "remove": [...]}
            # But that means in order to approximate upstream, behavior
            # which just overwrites the existing alias, we need to read
            # the bug's state first to know what string to remove. Which
            # we can't do, since we don't know the bug numbers at this point.
            # So fail for now.
            #
            # The API should provide {"set": [...]}
            # https://bugzilla.redhat.com/show_bug.cgi?id=1173114
            #
            # Implementation will go here when it's available
            pass

        if fixed_in is not None:
            adddict["cf_fixed_in"] = fixed_in
        if qa_whiteboard is not None:
            adddict["cf_qa_whiteboard"] = qa_whiteboard
        if devel_whiteboard is not None:
            adddict["cf_devel_whiteboard"] = devel_whiteboard
        if internal_whiteboard is not None:
            adddict["cf_internal_whiteboard"] = internal_whiteboard

        if sub_component:
            if not isinstance(sub_component, dict):
                component = listify(component)
                if not component:
                    raise ValueError("component must be specified if "
                        "specifying sub_component")
                sub_component = {component[0]: sub_component}
            adddict["sub_components"] = sub_component

        get_alias()

        return adddict


    #################
    # Query methods #
    #################

    @staticmethod
    def pre_translation(query):
        """
        Translates the query for possible aliases
        """
        old = query.copy()

        def split_comma(_v):
            if isinstance(_v, list):
                return _v
            return _v.split(",")

        if 'bug_id' in query:
            query['id'] = split_comma(query.pop('bug_id'))

        if 'component' in query:
            query['component'] = split_comma(query['component'])

        if 'include_fields' not in query and 'column_list' in query:
            query['include_fields'] = query.pop('column_list')

        if old != query:
            log.debug("RHBugzilla pretranslated query to: %s", query)

    @staticmethod
    def post_translation(query, bug):
        """
        Convert the results of getbug back to the ancient RHBZ value
        formats
        """
        ignore = query

        # RHBZ _still_ returns component and version as lists, which
        # deviates from upstream. Copy the list values to components
        # and versions respectively.
        if 'component' in bug and "components" not in bug:
            val = bug['component']
            bug['components'] = isinstance(val, list) and val or [val]
            bug['component'] = bug['components'][0]

        if 'version' in bug and "versions" not in bug:
            val = bug['version']
            bug['versions'] = isinstance(val, list) and val or [val]
            bug['version'] = bug['versions'][0]

        # sub_components isn't too friendly of a format, add a simpler
        # sub_component value
        if 'sub_components' in bug and 'sub_component' not in bug:
            val = bug['sub_components']
            bug['sub_component'] = ""
            if isinstance(val, dict):
                values = []
                for vallist in val.values():
                    values += vallist
                bug['sub_component'] = " ".join(values)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import tests
import tests.mockbackend
import tests.utils


##################################
# 'bugzilla modify' mock testing #
##################################

def test_modify(run_cli):
    # errors on missing args
    cmd = "bugzilla modify 123456"
    fakebz = tests.mockbackend.make_bz()
    out = run_cli(cmd, fakebz, expectfail=True)
    assert "additional arguments" in out

    # Modify basic
    cmd = "bugzilla modify 123456 1234567 "
    cmd += "--status ASSIGNED --component NEWCOMP "
    fakebz = tests.mockbackend.make_bz(
        bug_update_args="data/mockargs/test_modify1.txt",
        bug_update_return={})
    out = run_cli(cmd, fakebz)
    assert not out

    # Modify with lots of opts
    cmd = "bugzilla modify 123456 --component NEWCOMP "
    cmd += "--keyword +FOO --groups=-BAR --blocked =123456,445566 "
    cmd += "--flag=-needinfo,+somethingelse "
    cmd += "--whiteboard =foo --whiteboard =thisone "
    cmd += "--dupeid 555666 "
    cmd += "--comment 'some example comment' --private "
    fakebz = tests.mockbackend.make_bz(
        bug_update_args="data/mockargs/test_modify2.txt",
        bug_update_return={})
    out = run_cli(cmd, fakebz)
    assert not out

    # Modify with tricky opts hitting other API calls
    cmd = "bugzilla modify 1165434 "
    cmd += "--tags +addtag --tags=-rmtag "
    cmd += "--qa_whiteboard +yo-qa --qa_whiteboard=-foo "
    cmd += "--internal_whiteboard +internal-hey --internal_whiteboard +bar "
    cmd += "--devel_whiteboard +devel-duh --devel_whiteboard=-yay "
    fakebz = tests.mockbackend.make_bz(rhbz=True,
        bug_update_tags_args="data/mockargs/test_modify3-tags.txt",
        bug_update_tags_return={},
        bug_update_args="data/mockargs/test_modify3.txt",
        bug_update_return={},
        bug_get_args=None,
        bug_get_return="data/mockreturn/test_getbug_rhel.txt")
    out = run_cli(cmd, fakebz)
    assert not out

    # Modify hitting some rhbz paths
    cmd = "bugzilla modify 1165434 "
    cmd += "--fixed_in foofixedin "
    cmd += "--component lvm2 "
    cmd += "--sub-component some-sub-component"
    fakebz = tests.mockbackend.make_bz(rhbz=True,
        bug_update_args="data/mockargs/test_modify4.txt",
        bug_update_return={})
    out = run_cli(cmd, fakebz)
    assert not out

    # Modify with a slew of misc opt coverage
    cmd = "bugzilla modify 1165434 "
    cmd += "--assigned_to foo@example.com --qa_contact qa@example.com "
    cmd += "--product newproduct "
    cmd += "--blocked +1234 --blocked -1235 --blocked = "
    cmd += "--url https://example.com "
    cmd += "--cc=+bar@example.com --cc=-steve@example.com "
    cmd += "--dependson=+2234 --dependson=-2235 --dependson = "
    cmd += "--groups +foogroup "
    cmd += "--keywords +newkeyword --keywords=-byekeyword --keywords = "
    cmd += "--os windows --arch mips "
    cmd += "--priority high --severity low "
    cmd += "--summary newsummary --version 1.2.3 "
    cmd += "--reset-assignee --reset-qa-contact "
    cmd += "--alias fooalias "
    cmd += "--target_release 1.2.4 --target_milestone beta "
    cmd += "--devel_whiteboard =DEVBOARD --internal_whiteboard =INTBOARD "
    cmd += "--qa_whiteboard =QABOARD "
    cmd += "--comment-tag FOOTAG --field bar=foo "
    cmd += "--minor-update "
    fakebz = tests.mockbackend.make_bz(rhbz=True,
        bug_update_args="data/mockargs/test_modify5.txt",
        bug_update_return={})
    out = run_cli(cmd, fakebz)
    assert not out



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#

import os

import pytest

import tests
import tests.mockbackend


def test_api_attachments():
    # misc coverage testing for Bugzilla attachment APIs
    fakebz = tests.mockbackend.make_bz(
        bug_attachment_get_all_args=(
            "data/mockargs/test_attachments_getall1.txt"),
        bug_attachment_get_all_return={},
        bug_attachment_update_args=(
            "data/mockargs/test_attachments_update1.txt"),
        bug_attachment_update_return={},
        bug_attachment_get_args=(
            "data/mockargs/test_attachments_get1.txt"),
        bug_attachment_get_return=(
            "data/mockreturn/test_attach_get1.txt"),
        bug_attachment_create_args=(
            "data/mockargs/test_api_attachments_create1.txt"),
        bug_attachment_create_return={
            "attachments": {"123456": {}, "456789": []}},
    )

    # coverage for include/exclude handling
    fakebz.get_attachments([123456], None,
            include_fields=["foo"], exclude_fields="bar")

    # coverage for updateattachment
    fakebz.updateattachmentflags(None, "112233", "needinfo",
            value="foobar", is_patch=True)

    # coverage for openattachment
    fobj = fakebz.openattachment(502352)
    assert "Hooray" in str(fobj.read())

    # Error on bad input type
    with pytest.raises(TypeError):
        fakebz.attachfile([123456], None, "some desc")

    # Misc attachfile() pieces
    attachfile = os.path.dirname(__file__) + "/data/bz-attach-get1.txt"
    ret = fakebz.attachfile([123456], attachfile, "some desc",
            isprivate=True)
    ret.sort()
    assert ret == [123456, 456789]



#
# Copyright Red Hat, Inc. 2012
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#

"""
Test miscellaneous API bits
"""

import pytest

import bugzilla

import tests
import tests.mockbackend


def test_api_users():
    # Basic API testing of the users APIs
    user_ret = {'users': [
        {'can_login': True,
         'email': 'example1@example.com',
         'id': 1010101,
         'name': 'example1@example.com',
         'real_name': 'Mr. Example Man'},
        {'can_login': False,
         'email': 'example2@example.com',
         'id': 2222333,
         'name': 'example name',
         'real_name': 'Example real name',
         'saved_reports': [],
         'saved_searches': [],
         'groups': [
             {"id": 1, "name": "testgroup", "description": "somedesc"}
         ]},
    ]}

    # getusers and User testing
    fakebz = tests.mockbackend.make_bz(
        user_get_args="data/mockargs/test_api_users_get1.txt",
        user_get_return=user_ret,
        user_update_args="data/mockargs/test_api_users_update1.txt",
        user_update_return={})
    userobj = fakebz.getuser("example2@example.com")

    # Some userobj testing
    userobj.refresh()
    assert userobj.userid == 2222333
    assert userobj.email == "example2@example.com"
    assert userobj.name == "example name"
    assert userobj.can_login is False
    userobj.updateperms("rem", ["fedora_contrib"])

    # Catch a validation error
    with pytest.raises(bugzilla.BugzillaError):
        userobj.updateperms("badaction", ["newgroup"])

    # createuser tests
    fakebz = tests.mockbackend.make_bz(
        user_get_args="data/mockargs/test_api_users_get2.txt",
        user_get_return=user_ret,
        user_create_args="data/mockargs/test_api_users_create.txt",
        user_create_return={})
    userobj = fakebz.createuser("example1@example.com", "fooname", "foopass")
    assert userobj.email == "example1@example.com"


    # searchuser tests
    fakebz = tests.mockbackend.make_bz(
        user_get_args="data/mockargs/test_api_users_get3.txt",
        user_get_return=user_ret)
    userlist = fakebz.searchusers("example1@example.com")
    assert len(userlist) == 2



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.


class BugzillaError(Exception):
    """
    Error raised in the Bugzilla client code.
    """
    @staticmethod
    def get_bugzilla_error_string(exc):
        """
        Helper to return the bugzilla instance error message from an
        XMLRPC Fault, or any other exception type that's raised from bugzilla
        interaction
        """
        return getattr(exc, "faultString", str(exc))

    @staticmethod
    def get_bugzilla_error_code(exc):
        """
        Helper to return the bugzilla instance error code from an
        XMLRPC Fault, or any other exception type that's raised from bugzilla
        interaction
        """
        for propname in ["faultCode", "code"]:
            if hasattr(exc, propname):
                return getattr(exc, propname)
        return None

    def __init__(self, message, code=None):
        """
        :param code: The error code from the remote bugzilla instance. Only
            set if the error came directly from the remove bugzilla
        """
        self.code = code
        if self.code:
            message += " (code=%s)" % self.code
        Exception.__init__(self, message)



#!/usr/bin/env python3
#
# bugzilla - a commandline frontend for the python bugzilla module
#
# Copyright (C) 2007-2017 Red Hat Inc.
# Author: Will Woods <wwoods@redhat.com>
# Author: Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import argparse
import base64
import datetime
import errno
import json
import locale
from logging import getLogger, DEBUG, INFO, WARN, StreamHandler, Formatter
import os
import re
import socket
import sys
import tempfile
import urllib.parse
import xmlrpc.client

import requests.exceptions

import bugzilla


DEFAULT_BZ = 'https://bugzilla.redhat.com'

format_field_re = re.compile("%{([a-z0-9_]+)(?::([^}]*))?}")

log = getLogger(bugzilla.__name__)


################
# Util helpers #
################

def _is_unittest_debug():
    return bool(os.getenv("__BUGZILLA_UNITTEST_DEBUG"))


def open_without_clobber(name, *args):
    """
    Try to open the given file with the given mode; if that filename exists,
    try "name.1", "name.2", etc. until we find an unused filename.
    """
    fd = None
    count = 1
    orig_name = name
    while fd is None:
        try:
            fd = os.open(name, os.O_CREAT | os.O_EXCL, 0o666)
        except OSError as err:
            if err.errno == errno.EEXIST:
                name = "%s.%i" % (orig_name, count)
                count += 1
            else:  # pragma: no cover
                raise IOError(err.errno, err.strerror, err.filename) from None
    fobj = open(name, *args)
    if fd != fobj.fileno():
        os.close(fd)
    return fobj


def setup_logging(debug, verbose):
    handler = StreamHandler(sys.stderr)
    handler.setFormatter(Formatter(
        "[%(asctime)s] %(levelname)s (%(module)s:%(lineno)d) %(message)s",
        "%H:%M:%S"))
    log.addHandler(handler)

    if debug:
        log.setLevel(DEBUG)
    elif verbose:
        log.setLevel(INFO)
    else:
        log.setLevel(WARN)

    if _is_unittest_debug():
        log.setLevel(DEBUG)  # pragma: no cover


##################
# Option parsing #
##################

def _setup_root_parser():
    epilog = 'Try "bugzilla COMMAND --help" for command-specific help.'
    p = argparse.ArgumentParser(epilog=epilog)

    default_url = bugzilla.Bugzilla.get_rcfile_default_url()
    if not default_url:
        default_url = DEFAULT_BZ

    # General bugzilla connection options
    p.add_argument('--bugzilla', default=default_url,
            help="bugzilla URI. default: %s" % default_url)
    p.add_argument("--nosslverify", dest="sslverify",
                 action="store_false", default=True,
                 help="Don't error on invalid bugzilla SSL certificate")
    p.add_argument('--cert',
            help="client side certificate file needed by the webserver")

    p.add_argument('--login', action="store_true",
        help='Run interactive "login" before performing the '
             'specified command.')
    p.add_argument('--username', help="Log in with this username")
    p.add_argument('--password', help="Log in with this password")
    p.add_argument('--restrict-login', action="store_true",
                   help="The session (login token) will be restricted to "
                        "the current IP address.")

    p.add_argument('--ensure-logged-in', action="store_true",
        help="Raise an error if we aren't logged in to bugzilla. "
             "Consider using this if you are depending on "
             "cached credentials, to ensure that when they expire the "
             "tool errors, rather than subtly change output.")
    p.add_argument('--no-cache-credentials',
        action='store_false', default=True, dest='cache_credentials',
        help="Don't save any bugzilla cookies or tokens to disk, and "
             "don't use any pre-existing credentials.")

    p.add_argument('--cookiefile', default=None, help=argparse.SUPPRESS)
    p.add_argument('--tokenfile', default=None,
            help="token file to use for bugzilla authentication")

    p.add_argument('--verbose', action='store_true',
            help="give more info about what's going on")
    p.add_argument('--debug', action='store_true',
            help="output bunches of debugging info")
    p.add_argument('--version', action='version',
                   version=bugzilla.__version__)

    # Allow user to specify BZClass to initialize. Kinda weird for the
    # CLI, I'd rather people file bugs about this so we can fix our detection.
    # So hide it from the help output but keep it for back compat
    p.add_argument('--bztype', default='auto', help=argparse.SUPPRESS)

    return p


def _parser_add_output_options(p):
    outg = p.add_argument_group("Output format options")
    outg.add_argument('--full', action='store_const', dest='output',
            const='full', default='normal',
            help="output detailed bug info")
    outg.add_argument('-i', '--ids', action='store_const', dest='output',
            const='ids', help="output only bug IDs")
    outg.add_argument('-e', '--extra', action='store_const',
            dest='output', const='extra',
            help="output additional bug information "
                 "(keywords, Whiteboards, etc.)")
    outg.add_argument('--oneline', action='store_const', dest='output',
            const='oneline',
            help="one line summary of the bug (useful for scripts)")
    outg.add_argument('--json', action='store_const', dest='output',
            const='json', help="output contents in json format")
    outg.add_argument("--includefield", action="append",
            help="Pass the field name to bugzilla include_fields list. "
                 "Only the fields passed to include_fields are returned "
                 "by the bugzilla server. "
                 "This can be specified multiple times.")
    outg.add_argument("--extrafield", action="append",
            help="Pass the field name to bugzilla extra_fields list. "
                 "When used with --json this can be used to request "
                 "bugzilla to return values for non-default fields. "
                 "This can be specified multiple times.")
    outg.add_argument("--excludefield", action="append",
            help="Pass the field name to bugzilla exclude_fields list. "
                 "When used with --json this can be used to request "
                 "bugzilla to not return values for a field. "
                 "This can be specified multiple times.")
    outg.add_argument('--raw', action='store_const', dest='output',
            const='raw', help="raw output of the bugzilla contents. This "
            "format is unstable and difficult to parse. Use --json instead.")
    outg.add_argument('--outputformat',
            help="Print output in the form given. "
                 "You can use RPM-style tags that match bug "
                 "fields, e.g.: '%%{id}: %%{summary}'. See the man page "
                 "section 'Output options' for more details.")


def _parser_add_bz_fields(rootp, command):
    cmd_new = (command == "new")
    cmd_query = (command == "query")
    cmd_modify = (command == "modify")
    if cmd_new:
        comment_help = "Set initial bug comment/description"
    elif cmd_query:
        comment_help = "Search all bug comments"
    else:
        comment_help = "Add new bug comment"

    p = rootp.add_argument_group("Standard bugzilla options")

    p.add_argument('-p', '--product', help="Product name")
    p.add_argument('-v', '--version', help="Product version")
    p.add_argument('-c', '--component', help="Component name")
    p.add_argument('-t', '--summary', '--short_desc', help="Bug summary")
    p.add_argument('-l', '--comment', '--long_desc', help=comment_help)
    if not cmd_query:
        p.add_argument("--comment-tag", action="append",
                help="Comment tag for the new comment")
    p.add_argument("--sub-component", action="append",
        help="RHBZ sub component field")
    p.add_argument('-o', '--os', help="Operating system")
    p.add_argument('--arch', help="Arch this bug occurs on")
    p.add_argument('-x', '--severity', help="Bug severity")
    p.add_argument('-z', '--priority', help="Bug priority")
    p.add_argument('--alias', help='Bug alias (name)')
    p.add_argument('-s', '--status', '--bug_status',
        help='Bug status (NEW, ASSIGNED, etc.)')
    p.add_argument('-u', '--url', help="URL field")
    p.add_argument('-m', '--target_milestone', help="Target milestone")
    p.add_argument('--target_release', help="RHBZ Target release")

    p.add_argument('--blocked', action="append",
        help="Bug IDs that this bug blocks")
    p.add_argument('--dependson', action="append",
        help="Bug IDs that this bug depends on")
    p.add_argument('--keywords', action="append",
        help="Bug keywords")
    p.add_argument('--groups', action="append",
        help="Which user groups can view this bug")

    p.add_argument('--cc', action="append", help="CC list")
    p.add_argument('-a', '--assigned_to', '--assignee', help="Bug assignee")
    p.add_argument('-q', '--qa_contact', help='QA contact')
    if cmd_modify:
        p.add_argument("--minor-update", action="store_true",
                help="Request bugzilla to not send any "
                     "email about this change")

    if not cmd_new:
        p.add_argument('-f', '--flag', action='append',
            help="Bug flags state. Ex:\n"
                 "  --flag needinfo?\n"
                 "  --flag dev_ack+ \n"
                 "  clear with --flag needinfoX")
        p.add_argument("--tags", action="append",
                help="Tags/Personal Tags field.")

        p.add_argument('-w', "--whiteboard", '--status_whiteboard',
            action="append", help='Whiteboard field')
        p.add_argument("--devel_whiteboard", action="append",
            help='RHBZ devel whiteboard field')
        p.add_argument("--internal_whiteboard", action="append",
            help='RHBZ internal whiteboard field')
        p.add_argument("--qa_whiteboard", action="append",
            help='RHBZ QA whiteboard field')
        p.add_argument('-F', '--fixed_in',
            help="RHBZ 'Fixed in version' field")

    # Put this at the end, so it sticks out more
    p.add_argument('--field',
        metavar="FIELD=VALUE", action="append", dest="fields",
        help="Manually specify a bugzilla API field. FIELD is "
        "the raw name used by the bugzilla instance. For example, if your "
        "bugzilla instance has a custom field cf_my_field, do:\n"
        "  --field cf_my_field=VALUE")

    if not cmd_modify:
        _parser_add_output_options(rootp)


def _setup_action_new_parser(subparsers):
    description = ("Create a new bug report. "
        "--product, --component, --version, --summary, and --comment "
        "must be specified. "
        "Options that take multiple values accept comma separated lists, "
        "including --cc, --blocks, --dependson, --groups, and --keywords.")
    p = subparsers.add_parser("new", description=description)

    _parser_add_bz_fields(p, "new")
    g = p.add_argument_group("'new' specific options")
    g.add_argument('--private', action='store_true', default=False,
        help='Mark new comment as private')


def _setup_action_query_parser(subparsers):
    description = ("List bug reports that match the given criteria. "
        "Certain options can accept a comma separated list to query multiple "
        "values, including --status, --component, --product, --version, --id.")
    epilog = ("Note: querying via explicit command line options will only "
        "get you so far. See the --from-url option for a way to use powerful "
        "Web UI queries from the command line.")
    p = subparsers.add_parser("query",
        description=description, epilog=epilog)

    _parser_add_bz_fields(p, "query")

    g = p.add_argument_group("'query' specific options")
    g.add_argument('-b', '--id', '--bug_id',
        help="specify individual bugs by IDs, separated with commas")
    g.add_argument('-r', '--reporter',
        help="Email: search reporter email for given address")
    g.add_argument('--quicksearch',
        help="Search using bugzilla's quicksearch functionality.")
    g.add_argument('--savedsearch',
        help="Name of a bugzilla saved search. If you don't own this "
            "saved search, you must passed --savedsearch_sharer_id.")
    g.add_argument('--savedsearch-sharer-id',
        help="Owner ID of the --savedsearch. You can get this ID from "
            "the URL bugzilla generates when running the saved search "
            "from the web UI.")

    # Keep this at the end so it sticks out more
    g.add_argument('--from-url', metavar="WEB_QUERY_URL",
        help="Make a working query via bugzilla's 'Advanced search' web UI, "
             "grab the url from your browser (the string with query.cgi or "
             "buglist.cgi in it), and --from-url will run it via the "
             "bugzilla API. Don't forget to quote the string! "
             "This only works for Bugzilla 5 and Red Hat bugzilla")

    # Deprecated options
    p.add_argument('-E', '--emailtype', help=argparse.SUPPRESS)
    p.add_argument('--components_file', help=argparse.SUPPRESS)
    p.add_argument('-U', '--url_type',
            help=argparse.SUPPRESS)
    p.add_argument('-K', '--keywords_type',
            help=argparse.SUPPRESS)
    p.add_argument('-W', '--status_whiteboard_type',
            help=argparse.SUPPRESS)
    p.add_argument('--fixed_in_type', help=argparse.SUPPRESS)


def _setup_action_info_parser(subparsers):
    description = ("List products or component information about the "
        "bugzilla server.")
    p = subparsers.add_parser("info", description=description)

    x = p.add_mutually_exclusive_group(required=True)
    x.add_argument('-p', '--products', action='store_true',
            help='Get a list of products')
    x.add_argument('-c', '--components', metavar="PRODUCT",
            help='List the components in the given product')
    x.add_argument('-o', '--component_owners', metavar="PRODUCT",
            help='List components (and their owners)')
    x.add_argument('-v', '--versions', metavar="PRODUCT",
            help='List the versions for the given product')
    p.add_argument('--active-components', action="store_true",
            help='Only show active components. Combine with --components*')



def _setup_action_modify_parser(subparsers):
    usage = ("bugzilla modify [options] BUGID [BUGID...]\n"
        "Fields that take multiple values have a special input format.\n"
        "Append:    --cc=foo@example.com\n"
        "Overwrite: --cc==foo@example.com\n"
        "Remove:    --cc=-foo@example.com\n"
        "Options that accept this format: --cc, --blocked, --dependson,\n"
        "    --groups, --tags, whiteboard fields.")
    p = subparsers.add_parser("modify", usage=usage)

    _parser_add_bz_fields(p, "modify")

    g = p.add_argument_group("'modify' specific options")
    g.add_argument("ids", nargs="+", help="Bug IDs to modify")
    g.add_argument('-k', '--close', metavar="RESOLUTION",
        help='Close with the given resolution (WONTFIX, NOTABUG, etc.)')
    g.add_argument('-d', '--dupeid', metavar="ORIGINAL",
        help='ID of original bug. Implies --close DUPLICATE')
    g.add_argument('--private', action='store_true', default=False,
        help='Mark new comment as private')
    g.add_argument('--reset-assignee', action="store_true",
        help='Reset assignee to component default')
    g.add_argument('--reset-qa-contact', action="store_true",
        help='Reset QA contact to component default')


def _setup_action_attach_parser(subparsers):
    usage = """
bugzilla attach --file=FILE --desc=DESC [--type=TYPE] BUGID [BUGID...]
bugzilla attach --get=ATTACHID --getall=BUGID [--ignore-obsolete] [...]
bugzilla attach --type=TYPE BUGID [BUGID...]"""
    description = "Attach files or download attachments."
    p = subparsers.add_parser("attach", description=description, usage=usage)

    p.add_argument("ids", nargs="*", help="BUGID references")
    p.add_argument('-f', '--file', metavar="FILENAME",
            help='File to attach, or filename for data provided on stdin')
    p.add_argument('-d', '--description', '--summary',
            metavar="SUMMARY", dest='desc',
            help="A short summary of the file being attached")
    p.add_argument('-t', '--type', metavar="MIMETYPE",
            help="Mime-type for the file being attached")
    p.add_argument('-g', '--get', metavar="ATTACHID", action="append",
            default=[], help="Download the attachment with the given ID")
    p.add_argument("--getall", "--get-all", metavar="BUGID", action="append",
            default=[], help="Download all attachments on the given bug")
    p.add_argument('--ignore-obsolete', action="store_true",
        help='Do not download attachments marked as obsolete.')
    p.add_argument('-l', '--comment', '--long_desc',
            help="Add comment with attachment")
    p.add_argument('--private', action='store_true', default=False,
        help='Mark new comment as private')


def _setup_action_login_parser(subparsers):
    usage = 'bugzilla login [--api-key] [username [password]]'
    description = """Log into bugzilla and save a login cookie or token.
Note: These tokens are short-lived, and future Bugzilla versions will no
longer support token authentication at all. Please use a
~/.config/python-bugzilla/bugzillarc file with an API key instead, or
use 'bugzilla login --api-key' and we will save it for you."""
    p = subparsers.add_parser("login", description=description, usage=usage)
    p.add_argument('--api-key', action='store_true', default=False,
                   help='Prompt for and save an API key into bugzillarc, '
                        'rather than prompt for username and password.')
    p.add_argument("pos_username", nargs="?", help="Optional username ",
                   metavar="username")
    p.add_argument("pos_password", nargs="?", help="Optional password ",
                   metavar="password")


def setup_parser():
    rootparser = _setup_root_parser()
    subparsers = rootparser.add_subparsers(dest="command")
    subparsers.required = True
    _setup_action_new_parser(subparsers)
    _setup_action_query_parser(subparsers)
    _setup_action_info_parser(subparsers)
    _setup_action_modify_parser(subparsers)
    _setup_action_attach_parser(subparsers)
    _setup_action_login_parser(subparsers)
    return rootparser


####################
# Command routines #
####################

def _merge_field_opts(query, fields, parser):
    # Add any custom fields if specified
    for f in fields:
        try:
            f, v = f.split('=', 1)
            query[f] = v
        except Exception:
            parser.error("Invalid field argument provided: %s" % (f))


def _do_query(bz, opt, parser):
    q = {}

    # Parse preconstructed queries.
    u = opt.from_url
    if u:
        q = bz.url_to_query(u)

    if opt.components_file:
        # Components slurped in from file (one component per line)
        # This can be made more robust
        clist = []
        f = open(opt.components_file, 'r')
        for line in f.readlines():
            line = line.rstrip("\n")
            clist.append(line)
        opt.component = clist

    if opt.status:
        val = opt.status
        stat = val
        if val == 'ALL':
            # leaving this out should return bugs of any status
            stat = None
        elif val == 'DEV':
            # Alias for all development bug statuses
            stat = ['NEW', 'ASSIGNED', 'NEEDINFO', 'ON_DEV',
                'MODIFIED', 'POST', 'REOPENED']
        elif val == 'QE':
            # Alias for all QE relevant bug statuses
            stat = ['ASSIGNED', 'ON_QA', 'FAILS_QA', 'PASSES_QA']
        elif val == 'EOL':
            # Alias for EndOfLife bug statuses
            stat = ['VERIFIED', 'RELEASE_PENDING', 'CLOSED']
        elif val == 'OPEN':
            # non-Closed statuses
            stat = ['NEW', 'ASSIGNED', 'MODIFIED', 'ON_DEV', 'ON_QA',
                'VERIFIED', 'RELEASE_PENDING', 'POST']
        opt.status = stat

    # Convert all comma separated list parameters to actual lists,
    # which is what bugzilla wants
    # According to bugzilla docs, any parameter can be a list, but
    # let's only do this for options we explicitly mention can be
    # comma separated.
    for optname in ["severity", "id", "status", "component",
                    "priority", "product", "version"]:
        val = getattr(opt, optname, None)
        if not isinstance(val, str):
            continue
        setattr(opt, optname, val.split(","))

    include_fields = None
    if opt.output in ['raw', 'json']:
        # 'raw' always does a getbug() call anyways, so just ask for ID back
        include_fields = ['id']

    elif opt.outputformat:
        include_fields = []
        for fieldname, rest in format_field_re.findall(opt.outputformat):
            if fieldname == "whiteboard" and rest:
                fieldname = rest + "_" + fieldname
            elif fieldname == "flag":
                fieldname = "flags"
            elif fieldname == "cve":
                fieldname = ["keywords", "blocks"]
            elif fieldname == "__unicode__":
                # Needs to be in sync with bug.__unicode__
                fieldname = ["id", "status", "assigned_to", "summary"]

            flist = isinstance(fieldname, list) and fieldname or [fieldname]
            for f in flist:
                if f not in include_fields:
                    include_fields.append(f)

    if include_fields is not None:
        include_fields.sort()

    kwopts = {}
    if opt.product:
        kwopts["product"] = opt.product
    if opt.component:
        kwopts["component"] = opt.component
    if opt.sub_component:
        kwopts["sub_component"] = opt.sub_component
    if opt.version:
        kwopts["version"] = opt.version
    if opt.reporter:
        kwopts["reporter"] = opt.reporter
    if opt.id:
        kwopts["bug_id"] = opt.id
    if opt.summary:
        kwopts["short_desc"] = opt.summary
    if opt.comment:
        kwopts["long_desc"] = opt.comment
    if opt.cc:
        kwopts["cc"] = opt.cc
    if opt.assigned_to:
        kwopts["assigned_to"] = opt.assigned_to
    if opt.qa_contact:
        kwopts["qa_contact"] = opt.qa_contact
    if opt.status:
        kwopts["status"] = opt.status
    if opt.blocked:
        kwopts["blocked"] = opt.blocked
    if opt.dependson:
        kwopts["dependson"] = opt.dependson
    if opt.keywords:
        kwopts["keywords"] = opt.keywords
    if opt.keywords_type:
        kwopts["keywords_type"] = opt.keywords_type
    if opt.url:
        kwopts["url"] = opt.url
    if opt.url_type:
        kwopts["url_type"] = opt.url_type
    if opt.whiteboard:
        kwopts["status_whiteboard"] = opt.whiteboard
    if opt.status_whiteboard_type:
        kwopts["status_whiteboard_type"] = opt.status_whiteboard_type
    if opt.fixed_in:
        kwopts["fixed_in"] = opt.fixed_in
    if opt.fixed_in_type:
        kwopts["fixed_in_type"] = opt.fixed_in_type
    if opt.flag:
        kwopts["flag"] = opt.flag
    if opt.alias:
        kwopts["alias"] = opt.alias
    if opt.qa_whiteboard:
        kwopts["qa_whiteboard"] = opt.qa_whiteboard
    if opt.devel_whiteboard:
        kwopts["devel_whiteboard"] = opt.devel_whiteboard
    if opt.severity:
        kwopts["bug_severity"] = opt.severity
    if opt.priority:
        kwopts["priority"] = opt.priority
    if opt.target_release:
        kwopts["target_release"] = opt.target_release
    if opt.target_milestone:
        kwopts["target_milestone"] = opt.target_milestone
    if opt.emailtype:
        kwopts["emailtype"] = opt.emailtype
    if include_fields:
        kwopts["include_fields"] = include_fields
    if opt.quicksearch:
        kwopts["quicksearch"] = opt.quicksearch
    if opt.savedsearch:
        kwopts["savedsearch"] = opt.savedsearch
    if opt.savedsearch_sharer_id:
        kwopts["savedsearch_sharer_id"] = opt.savedsearch_sharer_id
    if opt.tags:
        kwopts["tags"] = opt.tags

    built_query = bz.build_query(**kwopts)
    if opt.fields:
        _merge_field_opts(built_query, opt.fields, parser)

    built_query.update(q)
    q = built_query

    if not q:  # pragma: no cover
        parser.error("'query' command requires additional arguments")
    return bz.query(q)


def _do_info(bz, opt):
    """
    Handle the 'info' subcommand
    """
    # All these commands call getproducts internally, so do it up front
    # with minimal include_fields for speed
    def _filter_components(compdetails):
        ret = {}
        for k, v in compdetails.items():
            if v.get("is_active", True):
                ret[k] = v
        return ret

    productname = (opt.components or opt.component_owners or opt.versions)
    fastcomponents = (opt.components and not opt.active_components)

    include_fields = ["name", "id"]
    if opt.components or opt.component_owners:
        include_fields += ["components.name"]
        if opt.component_owners:
            include_fields += ["components.default_assigned_to"]
        if opt.active_components:
            include_fields += ["components.is_active"]

    if opt.versions:
        include_fields += ["versions"]

    bz.refresh_products(names=productname and [productname] or None,
            include_fields=include_fields)

    if opt.products:
        for name in sorted([p["name"] for p in bz.getproducts()]):
            print(name)

    elif fastcomponents:
        for name in sorted(bz.getcomponents(productname)):
            print(name)

    elif opt.components:
        details = bz.getcomponentsdetails(productname)
        for name in sorted(_filter_components(details)):
            print(name)

    elif opt.versions:
        proddict = bz.getproducts()[0]
        for v in proddict['versions']:
            print(str(v["name"] or ''))

    elif opt.component_owners:
        details = bz.getcomponentsdetails(productname)
        for c in sorted(_filter_components(details)):
            print("%s: %s" % (c, details[c]['default_assigned_to']))


def _convert_to_outputformat(output):
    fmt = ""

    if output == "normal":
        fmt = "%{__unicode__}"

    elif output == "ids":
        fmt = "%{id}"

    elif output == 'full':
        fmt += "%{__unicode__}\n"
        fmt += "Component: %{component}\n"
        fmt += "CC: %{cc}\n"
        fmt += "Blocked: %{blocks}\n"
        fmt += "Depends: %{depends_on}\n"
        fmt += "%{comments}\n"

    elif output == 'extra':
        fmt += "%{__unicode__}\n"
        fmt += " +Keywords: %{keywords}\n"
        fmt += " +QA Whiteboard: %{qa_whiteboard}\n"
        fmt += " +Status Whiteboard: %{status_whiteboard}\n"
        fmt += " +Devel Whiteboard: %{devel_whiteboard}\n"

    elif output == 'oneline':
        fmt += "#%{bug_id} %{status} %{assigned_to} %{component}\t"
        fmt += "[%{target_milestone}] %{flags} %{cve}"

    else:  # pragma: no cover
        raise RuntimeError("Unknown output type '%s'" % output)

    return fmt


def _xmlrpc_converter(obj):
    if "DateTime" in str(obj.__class__):
        # xmlrpc DateTime object. Convert to date format that
        # bugzilla REST API outputs
        dobj = datetime.datetime.strptime(str(obj), '%Y%m%dT%H:%M:%S')
        return dobj.isoformat() + "Z"
    if "Binary" in str(obj.__class__):
        # xmlrpc Binary object. Convert to base64
        return base64.b64encode(obj.data).decode("utf-8")
    raise RuntimeError(
        "Unexpected JSON conversion class=%s" % obj.__class__)


def _format_output_json(buglist):
    out = {"bugs": [b.get_raw_data() for b in buglist]}
    s = json.dumps(out, default=_xmlrpc_converter, indent=2, sort_keys=True)
    print(s)


def _format_output_raw(buglist):
    for b in buglist:
        print("Bugzilla %s: " % b.bug_id)
        SKIP_NAMES = ["bugzilla"]
        for attrname in sorted(b.__dict__):
            if attrname in SKIP_NAMES:
                continue
            if attrname.startswith("_"):
                continue
            print("ATTRIBUTE[%s]: %s" % (attrname, b.__dict__[attrname]))
        print("\n\n")


def _bug_field_repl_cb(bz, b, matchobj):
    # whiteboard and flag allow doing
    #   %{whiteboard:devel} and %{flag:needinfo}
    # That's what 'rest' matches
    (fieldname, rest) = matchobj.groups()

    if fieldname == "whiteboard" and rest:
        fieldname = rest + "_" + fieldname

    if fieldname == "flag" and rest:
        val = b.get_flag_status(rest)

    elif fieldname in ["flags", "flags_requestee"]:
        tmpstr = []
        for f in getattr(b, "flags", []):
            requestee = f.get('requestee', "")
            if fieldname == "flags":
                requestee = ""
            if fieldname == "flags_requestee":
                if requestee == "":
                    continue
                tmpstr.append("%s" % requestee)
            else:
                tmpstr.append("%s%s%s" %
                        (f['name'], f['status'], requestee))

        val = ",".join(tmpstr)

    elif fieldname == "cve":
        cves = []
        for key in getattr(b, "keywords", []):
            # grab CVE from keywords and blockers
            if key.find("Security") == -1:
                continue
            for bl in b.blocks:
                cvebug = bz.getbug(bl)
                for cb in cvebug.alias:
                    if (cb.find("CVE") != -1 and
                        cb.strip() not in cves):
                        cves.append(cb)
        val = ",".join(cves)

    elif fieldname == "comments":
        val = ""
        for c in getattr(b, "comments", []):
            val += ("\n* %s - %s:\n%s\n" % (c['time'],
                     c.get("creator", c.get("author", "")), c['text']))

    elif fieldname == "external_bugs":
        val = ""
        for e in getattr(b, "external_bugs", []):
            url = e["type"]["full_url"].replace("%id%", e["ext_bz_bug_id"])
            if not val:
                val += "\n"
            val += "External bug: %s\n" % url

    elif fieldname == "__unicode__":
        val = b.__unicode__()
    else:
        val = getattr(b, fieldname, "")

    vallist = isinstance(val, list) and val or [val]
    val = ','.join([str(v or '') for v in vallist])

    return val


def _format_output(bz, opt, buglist):
    if opt.output in ['raw', 'json']:
        include_fields = None
        exclude_fields = None
        extra_fields = None

        if opt.includefield:
            include_fields = opt.includefield
        if opt.excludefield:
            exclude_fields = opt.excludefield
        if opt.extrafield:
            extra_fields = opt.extrafield

        buglist = bz.getbugs([b.bug_id for b in buglist],
                include_fields=include_fields,
                exclude_fields=exclude_fields,
                extra_fields=extra_fields)
        if opt.output == 'json':
            _format_output_json(buglist)
        if opt.output == 'raw':
            _format_output_raw(buglist)
        return

    for b in buglist:
        # pylint: disable=cell-var-from-loop
        def cb(matchobj):
            return _bug_field_repl_cb(bz, b, matchobj)
        print(format_field_re.sub(cb, opt.outputformat))


def _parse_triset(vallist, checkplus=True, checkminus=True, checkequal=True,
                  splitcomma=False):
    add_val = []
    rm_val = []
    set_val = None

    def make_list(v):
        if not v:
            return []
        if splitcomma:
            return v.split(",")
        return [v]

    for val in isinstance(vallist, list) and vallist or [vallist]:
        val = val or ""

        if val.startswith("+") and checkplus:
            add_val += make_list(val[1:])
        elif val.startswith("-") and checkminus:
            rm_val += make_list(val[1:])
        elif val.startswith("=") and checkequal:
            # Intentionally overwrite this
            set_val = make_list(val[1:])
        else:
            add_val += make_list(val)

    return add_val, rm_val, set_val


def _do_new(bz, opt, parser):
    # Parse options that accept comma separated list
    def parse_multi(val):
        return _parse_triset(val, checkplus=False, checkminus=False,
                             checkequal=False, splitcomma=True)[0]

    kwopts = {}
    if opt.blocked:
        kwopts["blocks"] = parse_multi(opt.blocked)
    if opt.cc:
        kwopts["cc"] = parse_multi(opt.cc)
    if opt.component:
        kwopts["component"] = opt.component
    if opt.dependson:
        kwopts["depends_on"] = parse_multi(opt.dependson)
    if opt.comment:
        kwopts["description"] = opt.comment
    if opt.groups:
        kwopts["groups"] = parse_multi(opt.groups)
    if opt.keywords:
        kwopts["keywords"] = parse_multi(opt.keywords)
    if opt.os:
        kwopts["op_sys"] = opt.os
    if opt.arch:
        kwopts["platform"] = opt.arch
    if opt.priority:
        kwopts["priority"] = opt.priority
    if opt.product:
        kwopts["product"] = opt.product
    if opt.severity:
        kwopts["severity"] = opt.severity
    if opt.summary:
        kwopts["summary"] = opt.summary
    if opt.url:
        kwopts["url"] = opt.url
    if opt.version:
        kwopts["version"] = opt.version
    if opt.assigned_to:
        kwopts["assigned_to"] = opt.assigned_to
    if opt.qa_contact:
        kwopts["qa_contact"] = opt.qa_contact
    if opt.sub_component:
        kwopts["sub_component"] = opt.sub_component
    if opt.alias:
        kwopts["alias"] = opt.alias
    if opt.comment_tag:
        kwopts["comment_tags"] = opt.comment_tag
    if opt.private:
        kwopts["comment_private"] = opt.private

    ret = bz.build_createbug(**kwopts)
    if opt.fields:
        _merge_field_opts(ret, opt.fields, parser)

    b = bz.createbug(ret)
    b.refresh()
    return [b]


def _do_modify(bz, parser, opt):
    bugid_list = [bugid for a in opt.ids for bugid in a.split(',')]

    add_wb, rm_wb, set_wb = _parse_triset(opt.whiteboard)
    add_devwb, rm_devwb, set_devwb = _parse_triset(opt.devel_whiteboard)
    add_intwb, rm_intwb, set_intwb = _parse_triset(opt.internal_whiteboard)
    add_qawb, rm_qawb, set_qawb = _parse_triset(opt.qa_whiteboard)

    add_blk, rm_blk, set_blk = _parse_triset(opt.blocked, splitcomma=True)
    add_deps, rm_deps, set_deps = _parse_triset(opt.dependson, splitcomma=True)
    add_key, rm_key, set_key = _parse_triset(opt.keywords)
    add_cc, rm_cc, ignore = _parse_triset(opt.cc,
                                          checkplus=False,
                                          checkequal=False)
    add_groups, rm_groups, ignore = _parse_triset(opt.groups,
                                                  checkequal=False,
                                                  splitcomma=True)
    add_tags, rm_tags, ignore = _parse_triset(opt.tags, checkequal=False)

    status = opt.status or None
    if opt.dupeid is not None:
        opt.close = "DUPLICATE"
    if opt.close:
        status = "CLOSED"

    flags = []
    if opt.flag:
        # Convert "foo+" to tuple ("foo", "+")
        for f in opt.flag:
            flags.append({"name": f[:-1], "status": f[-1]})

    update_opts = {}

    if opt.assigned_to:
        update_opts["assigned_to"] = opt.assigned_to
    if opt.comment:
        update_opts["comment"] = opt.comment
    if opt.private:
        update_opts["comment_private"] = opt.private
    if opt.component:
        update_opts["component"] = opt.component
    if opt.product:
        update_opts["product"] = opt.product
    if add_blk:
        update_opts["blocks_add"] = add_blk
    if rm_blk:
        update_opts["blocks_remove"] = rm_blk
    if set_blk is not None:
        update_opts["blocks_set"] = set_blk
    if opt.url:
        update_opts["url"] = opt.url
    if add_cc:
        update_opts["cc_add"] = add_cc
    if rm_cc:
        update_opts["cc_remove"] = rm_cc
    if add_deps:
        update_opts["depends_on_add"] = add_deps
    if rm_deps:
        update_opts["depends_on_remove"] = rm_deps
    if set_deps is not None:
        update_opts["depends_on_set"] = set_deps
    if add_groups:
        update_opts["groups_add"] = add_groups
    if rm_groups:
        update_opts["groups_remove"] = rm_groups
    if add_key:
        update_opts["keywords_add"] = add_key
    if rm_key:
        update_opts["keywords_remove"] = rm_key
    if set_key is not None:
        update_opts["keywords_set"] = set_key
    if opt.os:
        update_opts["op_sys"] = opt.os
    if opt.arch:
        update_opts["platform"] = opt.arch
    if opt.priority:
        update_opts["priority"] = opt.priority
    if opt.qa_contact:
        update_opts["qa_contact"] = opt.qa_contact
    if opt.severity:
        update_opts["severity"] = opt.severity
    if status:
        update_opts["status"] = status
    if opt.summary:
        update_opts["summary"] = opt.summary
    if opt.version:
        update_opts["version"] = opt.version
    if opt.reset_assignee:
        update_opts["reset_assigned_to"] = opt.reset_assignee
    if opt.reset_qa_contact:
        update_opts["reset_qa_contact"] = opt.reset_qa_contact
    if opt.close:
        update_opts["resolution"] = opt.close
    if opt.target_release:
        update_opts["target_release"] = opt.target_release
    if opt.target_milestone:
        update_opts["target_milestone"] = opt.target_milestone
    if opt.dupeid:
        update_opts["dupe_of"] = opt.dupeid
    if opt.fixed_in:
        update_opts["fixed_in"] = opt.fixed_in
    if set_wb and set_wb[0]:
        update_opts["whiteboard"] = set_wb and set_wb[0]
    if set_devwb and set_devwb[0]:
        update_opts["devel_whiteboard"] = set_devwb and set_devwb[0]
    if set_intwb and set_intwb[0]:
        update_opts["internal_whiteboard"] = set_intwb and set_intwb[0]
    if set_qawb and set_qawb[0]:
        update_opts["qa_whiteboard"] = set_qawb and set_qawb[0]
    if opt.sub_component:
        update_opts["sub_component"] = opt.sub_component
    if opt.alias:
        update_opts["alias"] = opt.alias
    if flags:
        update_opts["flags"] = flags
    if opt.comment_tag:
        update_opts["comment_tags"] = opt.comment_tag
    if opt.minor_update:
        update_opts["minor_update"] = opt.minor_update

    update = bz.build_update(**update_opts)

    # We make this a little convoluted to facilitate unit testing
    wbmap = {
        "whiteboard": (add_wb, rm_wb),
        "internal_whiteboard": (add_intwb, rm_intwb),
        "qa_whiteboard": (add_qawb, rm_qawb),
        "devel_whiteboard": (add_devwb, rm_devwb),
    }

    for k, v in wbmap.copy().items():
        if not v[0] and not v[1]:
            del(wbmap[k])

    if opt.fields:
        _merge_field_opts(update, opt.fields, parser)

    log.debug("update bug dict=%s", update)
    log.debug("update whiteboard dict=%s", wbmap)

    if not any([update, wbmap, add_tags, rm_tags]):
        parser.error("'modify' command requires additional arguments")

    if add_tags or rm_tags:
        ret = bz.update_tags(bugid_list,
            tags_add=add_tags, tags_remove=rm_tags)
        log.debug("bz.update_tags returned=%s", ret)
    if update:
        ret = bz.update_bugs(bugid_list, update)
        log.debug("bz.update_bugs returned=%s", ret)

    if not wbmap:
        return

    # Now for the things we can't blindly batch.
    # Being able to prepend/append to whiteboards, which are just
    # plain string values, is an old rhbz semantic that we try to maintain
    # here. This is a bit weird for traditional bugzilla API
    log.debug("Adjusting whiteboard fields one by one")
    for bug in bz.getbugs(bugid_list):
        update_kwargs = {}
        for wbkey, (add_list, rm_list) in wbmap.items():
            bugval = getattr(bug, wbkey) or ""
            for tag in add_list:
                if bugval:
                    bugval += " "
                bugval += tag

            for tag in rm_list:
                bugsplit = bugval.split()
                for t in bugsplit[:]:
                    if t == tag:
                        bugsplit.remove(t)
                bugval = " ".join(bugsplit)

            update_kwargs[wbkey] = bugval

        bz.update_bugs([bug.id], bz.build_update(**update_kwargs))


def _do_get_attach(bz, opt):
    data = {}

    def _process_attachment_data(_attlist):
        for _att in _attlist:
            data[_att["id"]] = _att

    if opt.getall:
        for attlist in bz.get_attachments(opt.getall, None)["bugs"].values():
            _process_attachment_data(attlist)
    if opt.get:
        _process_attachment_data(
            bz.get_attachments(None, opt.get)["attachments"].values())

    for attdata in data.values():
        is_obsolete = attdata.get("is_obsolete", None) == 1
        if opt.ignore_obsolete and is_obsolete:
            continue

        att = bz.openattachment_data(attdata)
        outfile = open_without_clobber(att.name, "wb")
        data = att.read(4096)
        while data:
            outfile.write(data)
            data = att.read(4096)
        print("Wrote %s" % outfile.name)


def _do_set_attach(bz, opt, parser):
    if not opt.ids:
        parser.error("Bug ID must be specified for setting attachments")

    if sys.stdin.isatty():
        if not opt.file:
            parser.error("--file must be specified")
        fileobj = open(opt.file, "rb")
    else:
        # piped input on stdin
        if not opt.desc:
            parser.error("--description must be specified if passing "
                         "file on stdin")

        fileobj = tempfile.NamedTemporaryFile(prefix="bugzilla-attach.")
        data = sys.stdin.read(4096)

        while data:
            fileobj.write(data.encode(locale.getpreferredencoding()))
            data = sys.stdin.read(4096)
        fileobj.seek(0)

    kwargs = {}
    if opt.file:
        kwargs["filename"] = os.path.basename(opt.file)
    if opt.type:
        kwargs["contenttype"] = opt.type
    if opt.type in ["text/x-patch"]:
        kwargs["ispatch"] = True
    if opt.comment:
        kwargs["comment"] = opt.comment
    if opt.private:
        kwargs["is_private"] = True
    desc = opt.desc or os.path.basename(fileobj.name)

    # Upload attachments
    for bugid in opt.ids:
        attid = bz.attachfile(bugid, fileobj, desc, **kwargs)
        print("Created attachment %i on bug %s" % (attid, bugid))


#################
# Main handling #
#################

def _make_bz_instance(opt):
    """
    Build the Bugzilla instance we will use
    """
    if opt.bztype != 'auto':
        log.info("Explicit --bztype is no longer supported, ignoring")

    cookiefile = None
    tokenfile = None
    use_creds = False
    if opt.cache_credentials:
        cookiefile = opt.cookiefile or -1
        tokenfile = opt.tokenfile or -1
        use_creds = True

    return bugzilla.Bugzilla(
        url=opt.bugzilla,
        cookiefile=cookiefile,
        tokenfile=tokenfile,
        sslverify=opt.sslverify,
        use_creds=use_creds,
        cert=opt.cert)


def _handle_login(opt, action, bz):
    """
    Handle all login related bits
    """
    is_login_command = (action == 'login')

    do_interactive_login = (is_login_command or
        opt.login or opt.username or opt.password)
    username = getattr(opt, "pos_username", None) or opt.username
    password = getattr(opt, "pos_password", None) or opt.password
    use_key = getattr(opt, "api_key", False)

    try:
        if use_key:
            bz.interactive_save_api_key()
        elif do_interactive_login:
            if bz.api_key:
                print("You already have an API key configured for %s" % bz.url)
                print("There is no need to cache a login token. Exiting.")
                sys.exit(0)
            print("Logging into %s" % urllib.parse.urlparse(bz.url)[1])
            bz.interactive_login(username, password,
                    restrict_login=opt.restrict_login)
    except bugzilla.BugzillaError as e:
        print(str(e))
        sys.exit(1)

    if opt.ensure_logged_in and not bz.logged_in:
        print("--ensure-logged-in passed but you aren't logged in to %s" %
            bz.url)
        sys.exit(1)

    if is_login_command:
        sys.exit(0)


def _main(unittest_bz_instance):
    parser = setup_parser()
    opt = parser.parse_args()
    action = opt.command
    setup_logging(opt.debug, opt.verbose)

    log.debug("Launched with command line: %s", " ".join(sys.argv))
    log.debug("Bugzilla module: %s", bugzilla)

    if unittest_bz_instance:
        bz = unittest_bz_instance
    else:
        bz = _make_bz_instance(opt)

    # Handle login options
    _handle_login(opt, action, bz)


    ###########################
    # Run the actual commands #
    ###########################

    if hasattr(opt, "outputformat"):
        if not opt.outputformat and opt.output not in ['raw', 'json', None]:
            opt.outputformat = _convert_to_outputformat(opt.output)

    buglist = []
    if action == 'info':
        _do_info(bz, opt)

    elif action == 'query':
        buglist = _do_query(bz, opt, parser)

    elif action == 'new':
        buglist = _do_new(bz, opt, parser)

    elif action == 'attach':
        if opt.get or opt.getall:
            if opt.ids:
                parser.error("Bug IDs '%s' not used for "
                    "getting attachments" % opt.ids)
            _do_get_attach(bz, opt)
        else:
            _do_set_attach(bz, opt, parser)

    elif action == 'modify':
        _do_modify(bz, parser, opt)
    else:  # pragma: no cover
        raise RuntimeError("Unexpected action '%s'" % action)

    # If we're doing new/query/modify, output our results
    if action in ['new', 'query']:
        _format_output(bz, opt, buglist)


def main(unittest_bz_instance=None):
    try:
        try:
            return _main(unittest_bz_instance)
        except (Exception, KeyboardInterrupt):
            log.debug("", exc_info=True)
            raise
    except KeyboardInterrupt:
        print("\nExited at user request.")
        sys.exit(1)
    except (xmlrpc.client.Fault, bugzilla.BugzillaError) as e:
        print("\nServer error: %s" % str(e))
        sys.exit(3)
    except requests.exceptions.SSLError as e:
        # Give SSL recommendations
        print("SSL error: %s" % e)
        print("\nIf you trust the remote server, you can work "
              "around this error with:\n"
              "  bugzilla --nosslverify ...")
        sys.exit(4)
    except (socket.error,
            requests.exceptions.HTTPError,
            requests.exceptions.ConnectionError,
            requests.exceptions.InvalidURL,
            xmlrpc.client.ProtocolError) as e:
        print("\nConnection lost/failed: %s" % str(e))
        sys.exit(2)


def cli():
    main()



#!/usr/bin/python3

import os
import unittest
from copy import deepcopy
import tempfile

from utils import getconfig
from utils import ConfigNotFoundException
from utils import ConfigMissingMandatoryFieldException
from utils import ConfigMalformedFieldException
from utils import ConfigDeprecatedValueException
from utils import SubmissionMethod

class ConfigTestCase(unittest.TestCase):
    
    validConfigInput = {
        "user": "some@user.com",
        "partner": "some_partner",
        "server": "https://bugzilla.example.com/xmlrpc.cgi",
        "partnergroup": "some_group",
        "api_key": "0123456789abcdef",
        "to": "some@list.com"
    }

    @staticmethod
    def writeConfigKeyValuePair(fp, key, value):
        fp.write(f"{key}={value}\n")

    @staticmethod
    def configFromDict(path, dic):
        with open(path, "w") as fp:
            for key in dic:
                ConfigTestCase.writeConfigKeyValuePair(fp, key, dic[key])

    def setUp(self):
        fd, self.configFile = tempfile.mkstemp()
        os.close(fd)

    def tearDown(self):
        os.unlink(self.configFile)

class BlankConfigTestCase(unittest.TestCase):
    def test_FailOnMissing(self):
        with self.assertRaises(ConfigNotFoundException) as context:
            getconfig("/invalid.path", False, False, False)

class ValidConfigTestCase(ConfigTestCase):
    def setUp(self):
        super().setUp()

    def test_AssertValidValues(self):
        configInput = deepcopy(self.validConfigInput)
        ConfigTestCase.configFromDict(self.configFile, configInput)
        result = getconfig(self.configFile, False, False, False)
        for key in configInput:
            res_key = key
            if res_key == "partnergroup":
                res_key = "group"
            self.assertEqual(result[res_key], configInput[key])

class InvalidConfigTestCase(ConfigTestCase):
    def setUp(self):
        super().setUp()

    def test_LegacyDefaultValues_RaiseException(self):
        defaultValues = {
            "user": "user@redhat.com",
            "partner": "partner-name",
            "partnergroup": "partner-group",
            "api_key": "api_key"
        }
        for defaultKey in defaultValues:
            configInput = deepcopy(self.validConfigInput)
            configInput[defaultKey] = defaultValues[defaultKey]
            ConfigTestCase.configFromDict(self.configFile, configInput)

            with self.assertRaises(ConfigDeprecatedValueException) as context:
                result = getconfig(self.configFile, False, False, False)

    def test_BlankMandatory_RaiseException(self):
        configInput = deepcopy(self.validConfigInput)
        del configInput["user"]
        ConfigTestCase.configFromDict(self.configFile, configInput)
        with self.assertRaises(ConfigMissingMandatoryFieldException) as context:
            result = getconfig(self.configFile, False, False, False)

        configInput = deepcopy(self.validConfigInput)
        del configInput["partnergroup"]
        ConfigTestCase.configFromDict(self.configFile, configInput)
        with self.assertRaises(ConfigMissingMandatoryFieldException) as context:
            result = getconfig(self.configFile, False, False, False)

        configInput = deepcopy(self.validConfigInput)
        del configInput["to"]
        del configInput["api_key"]
        ConfigTestCase.configFromDict(self.configFile, configInput)

        with self.assertRaises(ConfigMissingMandatoryFieldException) as context:
            result = getconfig(self.configFile, False, False, False)

    def test_MalformedServer_RaiseException(self):
        configInput = deepcopy(self.validConfigInput)
        configInput["server"] = "https://bugzilla.example.com/xmlrpc.foo"
        ConfigTestCase.configFromDict(self.configFile, configInput)

        with self.assertRaises(ConfigMalformedFieldException) as context:
            result = getconfig(self.configFile, False, False, False)

    def test_ExpectedMethods_Match(self):
        configInput = deepcopy(self.validConfigInput)
        ConfigTestCase.configFromDict(self.configFile, configInput)
        result = getconfig(self.configFile, False, False, False)

        exp_method = SubmissionMethod.NONE.value
        exp_method |= SubmissionMethod.BUGZILLA.value
        exp_method |= SubmissionMethod.MAILING_LIST.value
        self.assertEqual(result['method'], exp_method)

        configInput = deepcopy(self.validConfigInput)
        del configInput["server"]
        del configInput["api_key"]
        ConfigTestCase.configFromDict(self.configFile, configInput)
        result = getconfig(self.configFile, False, False, False)

        exp_method = SubmissionMethod.NONE.value
        exp_method |= SubmissionMethod.MAILING_LIST.value
        self.assertEqual(result['method'], exp_method)

        configInput = deepcopy(self.validConfigInput)
        del configInput["to"]
        ConfigTestCase.configFromDict(self.configFile, configInput)
        result = getconfig(self.configFile, False, False, False)

        exp_method = SubmissionMethod.NONE.value
        exp_method |= SubmissionMethod.BUGZILLA.value
        self.assertEqual(result['method'], exp_method)

    def test_PartnerOrGroupNoneValuesStripped(self):
        for key_list in [["partner"], ["group"], ["partner", "group"]]:
            configInput = deepcopy(self.validConfigInput)
            for key in key_list:
                key = key if key != "group" else "partnergroup"
                configInput[key] = "none"
            ConfigTestCase.configFromDict(self.configFile, configInput)
            result = getconfig(self.configFile, False, False, False)
            for key in key_list:
                self.assertEqual(result[key], "")

if __name__ == '__main__':
    unittest.main()



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import difflib
import getpass
import inspect
import io
import os
import pprint
import shlex
import sys

import bugzilla._cli

import tests


def get_funcname():
    # Return calling function name
    return inspect.stack()[1][3]


def tests_path(filename):
    testdir = os.path.dirname(__file__)
    if testdir not in filename:
        return os.path.join(testdir, filename)
    return filename


def monkeypatch_getpass(monkeypatch):
    monkeypatch.setattr(getpass, "getpass", input)


def sanitize_json(rawout):
    # py2.7 leaves trailing whitespace after commas. strip it so
    # tests pass on both python versions
    return "\n".join([line.rstrip() for line in rawout.splitlines()])


def open_functional_bz(bzclass, url, kwargs):
    bz = bzclass(url, **kwargs)

    if kwargs.get("force_rest", False):
        assert bz.is_rest() is True
    if kwargs.get("force_xmlrpc", False):
        assert bz.is_xmlrpc() is True

    # Set a request timeout of 60 seconds
    os.environ["PYTHONBUGZILLA_REQUESTS_TIMEOUT"] = "60"
    return bz


def diff_compare(inputdata, filename, expect_out=None):
    """Compare passed string output to contents of filename"""
    def _process(data):
        if isinstance(data, tuple) and len(data) == 1:
            data = data[0]
        if isinstance(data, (dict, tuple)):
            out = pprint.pformat(data, width=81)
        else:
            out = str(data)
        if not out.endswith("\n"):
            out += "\n"
        return out

    actual_out = _process(inputdata)

    if filename:
        filename = tests_path(filename)
        if not os.path.exists(filename) or tests.CLICONFIG.REGENERATE_OUTPUT:
            open(filename, "w").write(actual_out)
        expect_out = open(filename).read()
    else:
        expect_out = _process(expect_out)

    diff = "".join(difflib.unified_diff(expect_out.splitlines(1),
                                        actual_out.splitlines(1),
                                        fromfile=filename or "Manual input",
                                        tofile="Generated Output"))
    if diff:
        raise AssertionError("Conversion outputs did not match.\n%s" % diff)


def do_run_cli(capsys, monkeypatch,
               argvstr, bzinstance,
               expectfail=False, stdin=None):
    """
    Run bin/bugzilla.main() directly with passed argv
    """
    argv = shlex.split(argvstr)
    monkeypatch.setattr(sys, "argv", argv)
    if stdin:
        monkeypatch.setattr(sys, "stdin", io.StringIO(stdin))
    else:
        monkeypatch.setattr(sys.stdin, "isatty", lambda: True)

    ret = 0
    try:
        # pylint: disable=protected-access
        if bzinstance is None:
            bugzilla._cli.cli()
        else:
            bugzilla._cli.main(unittest_bz_instance=bzinstance)
    except SystemExit as sys_e:
        ret = sys_e.code

    out, err = capsys.readouterr()
    outstr = out + err

    if ret != 0 and not expectfail:
        raise RuntimeError("Command failed with %d\ncmd=%s\nout=%s" %
                           (ret, argvstr, outstr))
    if ret == 0 and expectfail:
        raise RuntimeError("Command succeeded but we expected success\n"
                           "ret=%d\ncmd=%s\nout=%s" %
                           (ret, argvstr, outstr))
    return outstr



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import tests
import tests.mockbackend
import tests.utils


################################
# 'bugzilla info' mock testing #
################################

def test_info(run_cli):
    funcname = tests.utils.get_funcname()
    argsprefix = "data/mockargs/%s_" % funcname
    cliprefix = "data/clioutput/%s_" % funcname

    prod_accessible = {'ids': [1, 7]}
    prod_get = {'products': [
        {'id': 1, 'name': 'Prod 1 Test'},
        {'id': 7, 'name': 'test-fake-product'}
    ]}

    # info --products
    fakebz = tests.mockbackend.make_bz(
        product_get_accessible_args=None,
        product_get_accessible_return=prod_accessible,
        product_get_args=argsprefix + "products.txt",
        product_get_return=prod_get)
    cmd = "bugzilla info --products"
    out = run_cli(cmd, fakebz)
    tests.utils.diff_compare(out, cliprefix + "products.txt")

    # info --versions
    prod_get_ver = {'products': [
        {'id': 7, 'name': 'test-fake-product',
         'versions': [
             {'id': 360, 'is_active': True, 'name': '7.1'},
             {'id': 123, 'is_active': True, 'name': 'fooversion!'},
         ]},
    ]}
    fakebz = tests.mockbackend.make_bz(
        product_get_args=argsprefix + "versions.txt",
        product_get_return=prod_get_ver)
    cmd = "bugzilla info --versions test-fake-product"
    out = run_cli(cmd, fakebz)
    tests.utils.diff_compare(out, cliprefix + "versions.txt")

    # info --components
    prod_get_comp_active = {'products': [
        {'id': 7, 'name': 'test-fake-product',
         'components': [
             {'is_active': True, 'name': 'backend/kernel'},
             {'is_active': True, 'name': 'client-interfaces'},
         ]},
    ]}
    cmd = "bugzilla info --components test-fake-product"
    fakebz = tests.mockbackend.make_bz(
        product_get_args=argsprefix + "components.txt",
        product_get_return=prod_get_comp_active)
    out = run_cli(cmd, fakebz)
    tests.utils.diff_compare(out, cliprefix + "components.txt")

    # info --components --active-components
    cmd = "bugzilla info --components test-fake-product --active-components"
    fakebz = tests.mockbackend.make_bz(
        product_get_args=argsprefix + "components-active.txt",
        product_get_return=prod_get_comp_active)
    out = run_cli(cmd, fakebz)
    tests.utils.diff_compare(out, cliprefix + "components-active.txt")

    # info --components_owners
    cmd = "bugzilla info --component_owners test-fake-product"
    prod_get_comp_owners = {'products': [
        {'id': 7, 'name': 'test-fake-product',
         'components': [
             {'default_assigned_to': 'Fake Guy',
              'name': 'client-interfaces'},
             {'default_assigned_to': 'ANother fake dude!',
              'name': 'configuration'},
         ]},
    ]}
    fakebz = tests.mockbackend.make_bz(
        product_get_args=argsprefix + "components-owners.txt",
        product_get_return=prod_get_comp_owners)
    out = run_cli(cmd, fakebz)
    tests.utils.diff_compare(out, cliprefix + "components-owners.txt")



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import configparser
import os
from logging import getLogger
import urllib.parse

from ._util import listify

log = getLogger(__name__)


def _parse_hostname(url):
    # If http://example.com is passed, netloc=example.com path=""
    # If just example.com is passed, netloc="" path=example.com
    parsedbits = urllib.parse.urlparse(url)
    return parsedbits.netloc or parsedbits.path


def _makedirs(path):
    if os.path.exists(os.path.dirname(path)):
        return
    os.makedirs(os.path.dirname(path), 0o700)


def _default_cache_location(filename):
    """
    Determine default location for passed xdg filename.
    example: ~/.cache/python-bugzilla/bugzillarc
    """
    return os.path.expanduser("~/.cache/python-bugzilla/%s" % filename)


class _BugzillaRCFile(object):
    """
    Helper class for interacting with bugzillarc files
    """
    @staticmethod
    def get_default_configpaths():
        paths = [
            '/etc/bugzillarc',
            '~/.bugzillarc',
            '~/.config/python-bugzilla/bugzillarc',
        ]
        return paths

    def __init__(self):
        self._cfg = None
        self._configpaths = None
        self.set_configpaths(None)

    def set_configpaths(self, configpaths):
        configpaths = [os.path.expanduser(p) for p in
                       listify(configpaths or [])]

        cfg = configparser.ConfigParser()
        read_files = cfg.read(configpaths)
        if read_files:
            log.info("Found bugzillarc files: %s", read_files)

        self._cfg = cfg
        self._configpaths = configpaths or []

    def get_configpaths(self):
        return self._configpaths[:]

    def get_default_url(self):
        """
        Grab a default URL from bugzillarc [DEFAULT] url=X
        """
        cfgurl = self._cfg.defaults().get("url", None)
        if cfgurl is not None:
            log.debug("bugzillarc: found cli url=%s", cfgurl)
            return cfgurl

    def parse(self, url):
        """
        Find the section for the passed URL domain, and return all the fields
        """
        section = ""
        log.debug("bugzillarc: Searching for config section matching %s", url)

        urlhost = _parse_hostname(url)
        for sectionhost in sorted(self._cfg.sections()):
            # If the section is just a hostname, make it match
            # If the section has a / in it, do a substring match
            if "/" not in sectionhost:
                if sectionhost == urlhost:
                    section = sectionhost
            elif sectionhost in url:
                section = sectionhost
            if section:
                log.debug("bugzillarc: Found matching section: %s", section)
                break

        if not section:
            log.debug("bugzillarc: No section found")
            return {}
        return dict(self._cfg.items(section))


    def save_api_key(self, url, api_key):
        """
        Save the API_KEY in the config file. We use the last file
        in the configpaths list, which is the one with the highest
        precedence.
        """
        configpaths = self.get_configpaths()
        if not configpaths:
            return None

        config_filename = configpaths[-1]
        section = _parse_hostname(url)
        cfg = configparser.ConfigParser()
        cfg.read(config_filename)

        if section not in cfg.sections():
            cfg.add_section(section)

        cfg.set(section, 'api_key', api_key.strip())

        _makedirs(config_filename)
        with open(config_filename, 'w') as configfile:
            cfg.write(configfile)

        return config_filename


class _BugzillaTokenCache(object):
    """
    Class for interacting with a .bugzillatoken cache file
    """
    @staticmethod
    def get_default_path():
        return _default_cache_location("bugzillatoken")

    def __init__(self):
        self._filename = None
        self._cfg = None

    def _get_domain(self, url):
        domain = urllib.parse.urlparse(url)[1]
        if domain and domain not in self._cfg.sections():
            self._cfg.add_section(domain)
        return domain

    def get_value(self, url):
        domain = self._get_domain(url)
        if domain and self._cfg.has_option(domain, 'token'):
            return self._cfg.get(domain, 'token')
        return None

    def set_value(self, url, value):
        if self.get_value(url) == value:
            return

        domain = self._get_domain(url)
        if value is None:
            self._cfg.remove_option(domain, 'token')
        else:
            self._cfg.set(domain, 'token', value)

        if self._filename:
            _makedirs(self._filename)
            with open(self._filename, 'w') as _cfg:
                log.debug("Saving to _cfg")
                self._cfg.write(_cfg)

    def get_filename(self):
        return self._filename

    def set_filename(self, filename):
        log.debug("Using tokenfile=%s", filename)
        cfg = configparser.ConfigParser()
        if filename:
            cfg.read(filename)
        self._filename = filename
        self._cfg = cfg



#!/usr/bin/env python
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

# update.py: Make changes to an existing bug

import time

import bugzilla

# public test instance of bugzilla.redhat.com. It's okay to make changes
URL = "bugzilla.stage.redhat.com"
bzapi = bugzilla.Bugzilla(URL)
if not bzapi.logged_in:
    print("This example requires cached login credentials for %s" % URL)
    bzapi.interactive_login()


# Similar to build_query, build_update is a helper function that handles
# some bugzilla version incompatibility issues. All it does is return a
# properly formatted dict(), and provide friendly parameter names.
# The param names map to those accepted by Bugzilla Bug.update:
# https://bugzilla.readthedocs.io/en/latest/api/core/v1/bug.html#update-bug
#
# Example bug: https://bugzilla.stage.redhat.com/show_bug.cgi?id=427301
# Don't worry, changing things here is fine, and won't send any email to
# users or anything. It's what bugzilla.stage.redhat.com is for!
bug = bzapi.getbug(427301)
print("Bug id=%s original summary=%s" % (bug.id, bug.summary))

update = bzapi.build_update(summary="new example summary %s" % time.time())
bzapi.update_bugs([bug.id], update)

# Call bug.refresh() to update its cached state
bug.refresh()
print("Bug id=%s new summary=%s" % (bug.id, bug.summary))


# Now let's add a comment
comments = bug.getcomments()
print("Bug originally has %d comments" % len(comments))

update = bzapi.build_update(comment="new example comment %s" % time.time())
bzapi.update_bugs([bug.id], update)

# refresh() actually isn't required here because comments are fetched
# on demand
comments = bug.getcomments()
print("Bug now has %d comments. Last comment=%s" % (len(comments),
    comments[-1]["text"]))


# The 'bug' object actually has some old convenience APIs for specific
# actions like commenting, and closing. However these aren't recommended:
# they encourage splitting up bug edits when really batching should be done
# as much as possible, not only to make your code quicker and save strain
# on the bugzilla instance, but also to avoid spamming bugzilla users with
# redundant email from two modifications that could have been batched.



# Copyright 2012,2018 Red Hat Inc.
# Author: Kushal Das <kdas@redhat.com>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.  See
# http://www.gnu.org/copyleft/gpl.html for the full text of the
# license.
#

"""
Helper functions for ksc.
"""

import os
import re
import sys
import time
import getpass
import subprocess
import locale

from enum import Enum

from bugzilla import Bugzilla, BugzillaError

import smtplib
from email.utils import formatdate
from email.mime.text import MIMEText

class SubmissionMethod(Enum):
    NONE = 0
    BUGZILLA = 1 << 0
    MAILING_LIST = 1 << 1

class ConfigNotFoundException(Exception):
    def __init__(self, filepath):
        message = f"{filepath}: Could not open config file."
        super().__init__(message)

class ConfigMissingMandatoryFieldException(Exception):
    def __init__(self, filepath, field, opt=""):
        opt = f" {opt}" if opt else opt
        if type(field) is str:
            message = f"{filepath}: Mandatory field `{field}' not found.{opt}"
        elif type(field) is list:
            message = f"{filepath}: At least one of the following fields " + \
                "must be set :" + ", ".join(map(lambda x: f"`{x}'", field))
        super().__init__(message)

class ConfigMalformedFieldException(Exception):
    def __init__(self, filepath, field, value, opt=""):
        opt = f" {opt}" if opt else opt
        message = f"{filepath}: Field `{field}' has an invalid value: " + \
            "`{value}'.{opt}"
        super().__init__(message)

class ConfigDeprecatedValueException(Exception):
    def __init__(self, filepath, field, value):
        message = f"{filepath}: Field `{field}' is set to a deprecated " + \
            "value: `{value}'. Please consult ksc manpage (man ksc)."
        super().__init__(message)

# stablelist directory
WHPATH = '/lib/modules'
# Module.symvers directory
SRCPATH = '/usr/src/kernels'

def query_user(query, max_tries=10, is_valid=lambda x: len(x) > 0):
    """
    Queries user for a value.

    :arg query:     query string
    :arg max_tries: maximal number of times user will be prompted to give a
                    valid reply (avoid cycling)
    :arg is_valid:  lambda function that determines if and when user supplied
                    input is valid

    :return         response     if valid
    :return         ""           if received max_tries invalid reponses
    :return         ""           if we couldn't read data from stdin
    """
    tries_left = max_tries
    response = ""
    while not is_valid(response):
        if tries_left < max_tries:
            if response == "":
                print("Empty response received. Please try again.")
            else:
                print("Option `%s' is invalid. Please try again." % response)

        if tries_left == 0:
            print("Reached maximum number of invalid responses.")
            return ""

        try:
            tries_left = tries_left - 1
            response = input(query)
        except EOFError:
            print("Reached early EOF.")
            return ""

    return response

def query_user_bool(query):
    """
    Queries user for a Y/N value

    :arg query:     query string
    :return         response     if valid
    :return         ""           if received max_tries invalid reponses
    :return         ""           if we couldn't read data from stdin
    """
    check_fx = lambda x: x.lower() in ['y', 'n']
    return query_user(query, is_valid=check_fx)

def get_release_name():
    if not os.path.isfile('/etc/redhat-release'):
        print('This tool needs to run on Red Hat Enterprise Linux')
        return None

    with open('/etc/redhat-release', 'r') as fptr:
        release = fptr.read().split(' ')
        if len(release) <= 6:
            print('This tool needs to run on Red Hat Enterprise Linux')
            return None
    for rel in release:
        if re.match("\d.\d+",rel):
            return rel
    print('This tool needs to run on Red Hat Enterprise Linux')
    return None

def read_list(arch, kabipath, verbose=False):
    """
    Reads a stablelist file and returns the symbols
    """
    result = []
    fpath = os.path.join(WHPATH, kabipath, "kabi_stablelist_%s" % arch)
    if not os.path.isfile(fpath):
        fpath = os.path.join(WHPATH, kabipath, "kabi_whitelist_%s" % arch)
    if not os.path.isfile(fpath):  # pragma: no cover
        print("File not found:", fpath)
        return [], False
    try:
        if verbose:  # pragma: no cover
            print("Reading %s" % fpath)
        fptr = open(fpath)
        for line in fptr.readlines():
            if line.startswith("["):
                continue
            result.append(line.strip("\n\t"))
        fptr.close()
    except IOError as err:  # pragma: no cover
        print(err)
        print("stablelist missing")

    return result, True


def read_total_list(symvers=None):
    """
    Reads total symbol list and returns the list
    """
    if not symvers:
        release = os.uname()[2]
        symvers = os.path.join(SRCPATH, release, "Module.symvers")
    if not os.path.isfile(symvers):  # pragma: no cover
        print("File not found:", symvers)
        print("Do you have current kernel-devel package installed?")
        sys.exit(1)
    result = []
    try:
        with open(symvers, "r") as fptr:
            for line in fptr.readlines():
                if line.startswith("["):
                    continue  # pragma: no cover
                result.append(line.split()[1])
    except IOError as err:  # pragma: no cover
        print(err)
        print("Missing all symbol list")
    return result


def run(command):
    """
    runs the given command
    """
    env = os.environ.copy()
    if "LANG" in env:
        env["LANG"] = 'C'
    ret = subprocess.Popen(command, shell=True, stdin=subprocess.PIPE,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                           close_fds=True, env=env)
    out, err = ret.communicate()
    if err:
        errs = err.decode(locale.getpreferredencoding()).split(':', 1)
        raise IOError(errs[1].strip() if len(errs) > 1 else err)
    return out.decode(locale.getpreferredencoding())

def is_valid_mail(email):
    return re.fullmatch(r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b', email)

def getconfig(path='/etc/ksc.conf', mock=False, require_partner=False, verbose=True):
    """
    Returns the bugzilla config
    """
    result = {}
    result['partner'] = ''
    result['method'] = SubmissionMethod.NONE.value

    if not os.path.isfile(path):
        path = '/etc/ksc.conf'
    try:
        fptr = open(path)
        lines = fptr.readlines()
        fptr.close()
    except Exception as err:
        raise ConfigNotFoundException(path)

    # To be deprecated in the future:
    result['bugzilla_enable'] = True

    cat = None
    for line in lines:
        if not line:
            continue
        m = re.match(r'^\[([^]]*)\]$', line)
        if m:
            cat = m.groups()[0]
        if line.startswith("user="):
            result["user"] = line[5:-1]
        elif line.startswith("partner="):
            result["partner"] = line[8:-1]
        elif line.startswith("server="):
            result["server"] = line[7:-1]
        elif line.startswith("partnergroup="):
            result["group"] = line[13:-1]
        elif line.startswith("api_key="):
            result["api_key"] = line[8:-1]
        elif line.startswith("to="):
            result["to"] = line[3:-1]
        elif line.startswith("smtp="):
            result["smtp"] = line[5:-1]
        elif line.startswith("enable="):
            result[f"{cat}_enable"] = bool(int(line[7:-1]))

    if 'user' not in result or not result['user']:
        raise ConfigMissingMandatoryFieldException(path, "user")

    if not is_valid_mail(result['user']):
        raise ConfigMalformedFieldException(path, "user", result['user'],
            "Please provide a valid e-mail address.")

    if 'partner' not in result or not result['partner']:
        raise ConfigMissingMandatoryFieldException(path, "partner")

    if 'group' not in result or not result['group']:
        raise ConfigMissingMandatoryFieldException(path, "group")

    if result['partner'] == "none":
        result['partner'] = ""

    if result['group'] == "none":
        result['group'] = ""

    mandatory_set = False
    if 'to' in result and result['to']:
        if not is_valid_mail(result['to']):
            raise ConfigMalformedFieldException(path, "to", result['to'],
                "Please provide a valid e-mail address.")
        result['method'] |= SubmissionMethod.MAILING_LIST.value
        mandatory_set = True

    if 'api_key' in result and result['api_key']:
        if 'server' not in result or not result['server']:
            raise ConfigMissingMandatoryFieldException(path, "user",
                "Field `api_key' requires `server' field, which was not set.")
        if not result['server'].endswith('xmlrpc.cgi'):
            raise ConfigMalformedFieldException(path, "server", result['server'],
                "Please provide a valid RPC URL, e.g., " +
                "`https://bugzilla.redhat.com/xmlrpc.cgi'.")
        result['method'] |= SubmissionMethod.BUGZILLA.value
        mandatory_set = True

    if not mandatory_set:
        raise ConfigMissingMandatoryFieldException(path, ["to", "api_key"])

    # ksc deprecated default values, ensure that users do not try to file a bug
    # using legacy configuration's default values
    deprecatedValues = {
        "user": "user@redhat.com",
        "partner": "partner-name",
        "group": "partner-group",
        "api_key": "api_key"
    }

    for key in deprecatedValues:
        if key not in result:
            continue
        if not result[key]:
            continue
        if result[key] != deprecatedValues[key]:
            continue
        raise ConfigDeprecatedValueException(path, key, result[key])

    if not (result['method'] & SubmissionMethod.BUGZILLA.value):
        conf["bugzilla_enable"] = False

    if not (result['method'] & SubmissionMethod.MAILING_LIST.value):
        conf["mailing_list_enable"] = False

    return result

def submit_stable(filename, arch, mock=False, path='/etc/ksc.conf',
              releasename='9.0', module=None):
    return submit(filename, arch, mock, path, releasename, module,
            "kabi-stablelists")

def submit_notif(filename, arch, mock=False, path='/etc/ksc.conf',
              releasename='9.0', module=None):
    return submit(filename, arch, mock, path, releasename, module,
            "kabi-notificationlists")

def submit(filename, arch, mock, path, releasename, module, subcomponent,
        require_partner = False):
    if mock:
        print("Using local config file data/ksc.conf")
        path = './data/ksc.conf'

    try:
        conf = getconfig(path, mock, require_partner)
    except Exception as err:
        print("Problem in parsing the configuration.")
        print(err)
        return

    if conf["bugzilla_enable"]:
        print("Bugzilla: enabled")
        createbug(filename, arch, mock, conf, releasename, module,
                  subcomponent, require_partner)
    else:
        print("Bugzilla: disabled")

    if conf["mailing_list_enable"]:
        print("Mailing list: enabled")
        sendmail(filename, arch, mock, conf, releasename, module,
                 subcomponent, require_partner)
    else:
        print("Mailing list: disabled")

def get_major_release(releasename):
    centos = releasename[1]
    releasename = releasename[0]

    major = releasename.split(".")
    if len(major) < 2 or not releasename.split(".")[0].isnumeric():
        return None, None

    return int(releasename.split(".")[0]), centos

def sendmail(filename, arch, mock, conf, releasename, module, subcomponent,
        require_partner = False):
    """
    Email ksc report.
    """

    major, centos = get_major_release(releasename)
    if not major:
        print("Invalid releasename: Mail not sent.")
        return

    body  = f"Product:  Red Hat Enterprise Linux {major}\n"
    body += f"Release:  Centos Stream\n"
    body += f"Platform: {arch}\n"

    if 'group' in conf and conf['group'] != 'partner-group':
        body += f"Partner Group: {conf['group']}\n"

    if 'partner' in conf:
        body += f"Partner: {conf['partner']}\n"

    body += "\n"
    body += str(module) + "\n"
    body += "\n"
    body += "---\n"
    body += "\n"

    with open(filename, "r") as fp:
        for line in fp:
            body += f"{line}\n"

    msg = MIMEText(body)
    msg['To'] = conf['to']
    msg['From'] = conf['user']
    msg['Date'] = formatdate(localtime=True)
    msg['Subject'] = f"{subcomponent} subscription"

    smtp = smtplib.SMTP(conf['smtp'])
    smtp.sendmail(conf['user'], conf['to'], msg.as_string())
    smtp.close()

    print("Mail sent.")

def createbug(filename, arch, mock, conf, releasename, module, subcomponent,
        require_partner = False):
    """
    Opens a bug in the Bugzilla
    """

    major, centos = get_major_release(releasename)
    if not major:
        print("Invalid releasename: Bug not created")
        return

    bughash = {'product': f"Red Hat Enterprise Linux {major}"}

    bughash["component"] = 'kernel'
    bughash["summary"] = "kABI Symbol Usage"
    bughash["version"] = releasename[0] if not centos else "CentOS Stream"
    bughash["platform"] = arch
    bughash["severity"] = "medium"
    bughash["priority"] = "medium"
    bughash["description"] = "Creating the bug to attach the symbol " + \
                             "usage details."
    bughash["qa_contact"] = "kernel-qe@redhat.com"

    if module:
        bughash["summary"] += " ({})".format(str(module))

    if 'group' in conf and conf['group']:
        bughash["groups"] = [conf['group']]

    if 'api_key' in conf and conf['api_key'] != 'api_key':
        bughash["Bugzilla_api_key"] = conf["api_key"]
    else:
        bughash["Bugzilla_login"] = conf["user"]
        bughash["Bugzilla_password"] = conf["password"]

    if conf["partner"]:
        bughash["cf_partner"] = [conf["partner"], ]
    else:
        if require_partner:
            print("You must provide a valid non-empty Partner field when using -s.")
            sys.exit(1)
        if query_user_bool("You have provided blank partner field. " \
                "This will file your request publicly. Proceed? [y/N]: ") != 'y':
            print("ksc-report.txt not uploaded. Terminating...")
            sys.exit(1)

    bughash["keywords"] = ["Tracking"]

    try:
        if 'api_key' in conf and conf['api_key'] != 'api_key':
            bz = Bugzilla(
                url=conf['server'],
                api_key=conf["api_key"]
            )
        else:
            bz = Bugzilla(
                url=conf['server'],
                user=conf["user"],
                password=conf["password"]
            )
    except BugzillaError as err:
        print("Bug not submitted. %s" % err)
        if not mock:
            sys.exit(1)

    if not mock:  # pragma: no cover
        print("Creating a new bug")

    bughash["sub_component"] = subcomponent

    # As it is as yet unclear whether the new sub_component will be
    # set up at the time of deployment, attemp to file with the old
    # sub_component as well.
    # If kabi-stablelists does not exist, an attempt to createbug
    # will cause an xmlrpc.client.Fault exception.
    try:
        trycreatebug(filename, mock, bughash, conf, bz)
    except Exception as err:  # pragma: no cover
        try:
            bughash["sub_component"] = 'kabi-whitelists'
            trycreatebug(filename, mock, bughash, conf, bz)
        except Exception as err:  # pragma: no cover
            print ("Could not create bug. %s" % err)
            if not mock:
                sys.exit(1)

def trycreatebug(filename, mock, bughash, conf, bz):

    bugid = 0

    if "groups" in bughash:
        ret = bz.build_createbug(
            product=bughash['product'],
            component=bughash['component'],
            sub_component=bughash['sub_component'],
            summary=bughash['summary'],
            version=bughash['version'],
            platform=bughash['platform'],
            qa_contact=bughash['qa_contact'],
            severity=bughash['severity'],
            priority=bughash['priority'],
            description=bughash['description'],
            groups=bughash['groups'],
            keywords=bughash['keywords']
        )
    else:
        ret = bz.build_createbug(
            product=bughash['product'],
            component=bughash['component'],
            sub_component=bughash['sub_component'],
            summary=bughash['summary'],
            version=bughash['version'],
            platform=bughash['platform'],
            qa_contact=bughash['qa_contact'],
            severity=bughash['severity'],
            priority=bughash['priority'],
            description=bughash['description'],
            keywords=bughash['keywords']
        )

    if "cf_partner" in bughash and bughash['cf_partner']:
        ret['cf_partner'] = bughash['cf_partner']

    bug = bz.createbug(ret)

    bugid = bug.id

    if not mock:  # pragma: no cover
        print("Bug URL %s/show_bug.cgi?id=%s" % \
              (conf['server'][:-11], bugid))
        print("Attaching the report")

    dhash = {}
    dhash["filename"] = "ksc-result.txt"
    dhash["contenttype"] = "text/plain"
    desc = "kABI symbol usage."

    for _ in range(3):
        with open(filename, "r") as fptr:
            attachment_id = bz.attachfile(bugid, fptr, desc, **dhash)

        if not mock:  # pragma: no cover
            if not attachment_id:
                time.sleep(1)
            else:
                print("Attached successfully as %s on bug %s" % (attachment_id, bugid))
                break
    else:
        print("Failed to attach symbol usage result")
        sys.exit()

    return bugid



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.


def listify(val):
    """Ensure that value is either None or a list, converting single values
    into 1-element lists"""
    if val is None:
        return val
    if isinstance(val, list):
        return val
    return [val]



#!/usr/bin/env python
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

# getbug.py: Simple demonstration of connecting to bugzilla, fetching
#            a bug, and printing some details.

import pprint

import bugzilla

# public test instance of bugzilla.redhat.com. It's okay to make changes
URL = "bugzilla.stage.redhat.com"

bzapi = bugzilla.Bugzilla(URL)

# getbug() is just a simple wrapper around getbugs(), which takes a list
# IDs, if you need to fetch multiple
#
# Example bug: https://bugzilla.stage.redhat.com/show_bug.cgi?id=427301
bug = bzapi.getbug(427301)
print("Fetched bug #%s:" % bug.id)
print("  Product   = %s" % bug.product)
print("  Component = %s" % bug.component)
print("  Status    = %s" % bug.status)
print("  Resolution= %s" % bug.resolution)
print("  Summary   = %s" % bug.summary)

# Just check dir(bug) for other attributes, or check upstream bugzilla
# Bug.get docs for field names:
# https://bugzilla.readthedocs.io/en/latest/api/core/v1/bug.html#get-bug

# comments must be fetched separately on stock bugzilla. this just returns
# a raw dict with all the info.
comments = bug.getcomments()
print("\nLast comment data:\n%s" % pprint.pformat(comments[-1]))

# getcomments is just a wrapper around bzapi.get_comments(), which can be
# used for bulk comments fetching



#!/usr/bin/env python
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

# create.py: Create a new bug report

import time

import bugzilla

# public test instance of bugzilla.redhat.com.
#
# Don't worry, changing things here is fine, and won't send any email to
# users or anything. It's what bugzilla.stage.redhat.com is for!
URL = "bugzilla.stage.redhat.com"
bzapi = bugzilla.Bugzilla(URL)
if not bzapi.logged_in:
    print("This example requires cached login credentials for %s" % URL)
    bzapi.interactive_login()


# Similar to build_query, build_createbug is a helper function that handles
# some bugzilla version incompatibility issues. All it does is return a
# properly formatted dict(), and provide friendly parameter names.
# The argument names map to those accepted by Bugzilla Bug.create:
# https://bugzilla.readthedocs.io/en/latest/api/core/v1/bug.html#create-bug
#
# The arguments specified here are mandatory, but there are many other
# optional ones like op_sys, platform, etc. See the docs
createinfo = bzapi.build_createbug(
    product="Fedora",
    version="rawhide",
    component="python-bugzilla",
    summary="new example python-bugzilla bug %s" % time.time(),
    description="This is comment #0 of an example bug created by "
                "the python-bugzilla.git examples/create.py script.")

newbug = bzapi.createbug(createinfo)
print("Created new bug id=%s url=%s" % (newbug.id, newbug.weburl))



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from logging import getLogger

import requests

log = getLogger(__name__)


class _BackendBase(object):
    """
    Backends are thin wrappers around the different bugzilla API paradigms
    (XMLRPC, REST). This base class defines the public API for the rest of
    the code, but this is all internal to the library.
    """
    def __init__(self, url, bugzillasession):
        self._url = url
        self._bugzillasession = bugzillasession


    @staticmethod
    def probe(url):
        try:
            requests.head(url).raise_for_status()
            return True  # pragma: no cover
        except Exception as e:
            log.debug("Failed to probe url=%s : %s", url, str(e))
        return False


    #################
    # Internal APIs #
    #################

    def get_xmlrpc_proxy(self):
        """
        Provides the raw XMLRPC proxy to API users of Bugzilla._proxy
        """
        raise NotImplementedError()

    def is_rest(self):
        """
        :returns: True if this is the REST backend
        """
        return False

    def is_xmlrpc(self):
        """
        :returns: True if this is the XMLRPC backend
        """
        return False


    ######################
    # Bugzilla info APIs #
    ######################

    def bugzilla_version(self):
        """
        Fetch bugzilla version string
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/bugzilla.html#version
        """
        raise NotImplementedError()


    #######################
    # Bug attachment APIs #
    #######################

    def bug_attachment_get(self, attachment_ids, paramdict):
        """
        Fetch bug attachments IDs. One part of:
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/attachment.html#get-attachment
        """
        raise NotImplementedError()

    def bug_attachment_get_all(self, bug_ids, paramdict):
        """
        Fetch all bug attachments IDs. One part of
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/attachment.html#get-attachment
        """
        raise NotImplementedError()

    def bug_attachment_create(self, bug_ids, data, paramdict):
        """
        Create a bug attachment
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/attachment.html#create-attachment

        :param data: raw Bytes data of the attachment to attach. API will
            encode this correctly if you pass it in and 'data' is not in
            paramdict.
        """
        raise NotImplementedError()

    def bug_attachment_update(self, attachment_ids, paramdict):
        """
        Update a bug attachment
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/attachment.html#update-attachment
        """
        raise NotImplementedError()


    ############
    # bug APIs #
    ############

    def bug_comments(self, bug_ids, paramdict):
        """
        Fetch bug comments
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/comment.html#get-comments
        """
        raise NotImplementedError()

    def bug_create(self, paramdict):
        """
        Create a new bug
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/bug.html#create-bug
        """
        raise NotImplementedError()

    def bug_fields(self, paramdict):
        """
        Query available bug field values
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/field.html#fields
        """
        raise NotImplementedError()

    def bug_get(self, bug_ids, aliases, paramdict):
        """
        Lookup bug data by ID
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/bug.html#get-bug
        """
        raise NotImplementedError()

    def bug_history(self, bug_ids, paramdict):
        """
        Lookup bug history
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/bug.html#bug-history
        """
        raise NotImplementedError()

    def bug_search(self, paramdict):
        """
        Search/query bugs
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/bug.html#search-bugs
        """
        raise NotImplementedError()

    def bug_update(self, bug_ids, paramdict):
        """
        Update bugs
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/bug.html#update-bug
        """
        raise NotImplementedError()

    def bug_update_tags(self, bug_ids, paramdict):
        """
        Update bug tags
        https://www.bugzilla.org/docs/4.4/en/html/api/Bugzilla/WebService/Bug.html#update_tags
        """
        raise NotImplementedError()


    ##################
    # Component APIs #
    ##################

    def component_create(self, paramdict):
        """
        Create component
        https://bugzilla.readthedocs.io/en/latest/api/core/v1/component.html#create-component
        """
        raise NotImplementedError()

    def component_update(self, paramdict):
        """
        Update component
        https://bugzilla.readthedocs.io/en/latest/api/core/v1/component.html#update-component
        """
        raise NotImplementedError()


    ###############################
    # ExternalBugs extension APIs #
    ###############################

    def externalbugs_add(self, paramdict):
        """
        https://bugzilla.redhat.com/docs/en/html/integrating/api/Bugzilla/Extension/ExternalBugs/WebService.html#add-external-bug
        """
        raise NotImplementedError()

    def externalbugs_update(self, paramdict):
        """
        https://bugzilla.redhat.com/docs/en/html/integrating/api/Bugzilla/Extension/ExternalBugs/WebService.html#update-external-bug
        """
        raise NotImplementedError()

    def externalbugs_remove(self, paramdict):
        """
        https://bugzilla.redhat.com/docs/en/html/integrating/api/Bugzilla/Extension/ExternalBugs/WebService.html#remove-external-bug
        """
        raise NotImplementedError()


    ##############
    # Group APIs #
    ##############

    def group_get(self, paramdict):
        """
        https://bugzilla.readthedocs.io/en/latest/api/core/v1/group.html#get-group
        """
        raise NotImplementedError()


    ################
    # Product APIs #
    ################

    def product_get(self, paramdict):
        """
        Fetch product details
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/product.html#get-product
        """
        raise NotImplementedError()

    def product_get_accessible(self):
        """
        List accessible products
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/product.html#list-products
        """
        raise NotImplementedError()

    def product_get_enterable(self):
        """
        List enterable products
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/product.html#list-products
        """
        raise NotImplementedError()

    def product_get_selectable(self):
        """
        List selectable products
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/product.html#list-products
        """
        raise NotImplementedError()


    #############
    # User APIs #
    #############

    def user_create(self, paramdict):
        """
        Create user
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/user.html#create-user
        """
        raise NotImplementedError()

    def user_get(self, paramdict):
        """
        Get user info
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/user.html#get-user
        """
        raise NotImplementedError()

    def user_login(self, paramdict):
        """
        Log in to bugzilla
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/user.html#login
        """
        raise NotImplementedError()

    def user_logout(self):
        """
        Log out of bugzilla
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/user.html#logout
        """
        raise NotImplementedError()

    def user_update(self, paramdict):
        """
        Update user
        http://bugzilla.readthedocs.io/en/latest/api/core/v1/user.html#update-user
        """
        raise NotImplementedError()
